# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stats test base to be shared between TF and JAX AQTp ops.

Since both TF and JAX AQTp are supposed to have the exactly same logics and
APIs for all operations, it is desired to share their test cases. It will help
to make sure that both of them behave the same consistently and to maintain
and modify their test cases much easier whenever both the libraries evolve.
"""

from absl import flags
from absl.testing import parameterized
from aqt.common import aqt_config
from aqt.test import aqt_test_shared_base
import numpy as np
import tensorflow.compat.v1 as tf


def f32(x):
  """Creates a float32 numpy array."""
  return np.array(x, dtype=np.float32)


class StatsTest(tf.test.TestCase, parameterized.TestCase):
  """Base class for testing Stats.

  As TF and JAX AQTp have the same API and logic for Stats class, all testing
  cases are put in this base class and are shared between TF and JAX through
  inheritance. On TF and JAX AQTp sides, only few methods which are not
  framework agnostic should be implemented.
  """

  _stats = None

  def set_stats(self, data_shape, config):
    """Sets self._stats with given data shape and stats config.

    Should be implemented in TF and JAX AQTp tests with their own AQT ops.

    Args:
      data_shape: Shape of data for which statistics are collected.
      config: Configuration fot Stats.
    """
    raise NotImplementedError

  def update(self, sample, weight):
    """Updates self._stats with a given tensor and its weight."""

    raise NotImplementedError

  @property
  def stats(self):
    if self._stats is None:
      raise ValueError("set_stats() is not called")
    return self._stats

  @property
  def stats_state(self):
    if self._stats_state is None:
      raise ValueError("set_stats() is not called")
    return self._stats_state

  def get_sum_of_ones(self):
    raise NotImplementedError

  def get_sum_of_vals(self):
    raise NotImplementedError

  def get_max_of_abs_vals(self):
    raise NotImplementedError

  def get_sum_of_l1_vals(self):
    raise NotImplementedError

  def get_sum_of_lp_vals(self):
    raise NotImplementedError

  def set_ema_update_count(self, ema_update_count):
    raise NotImplementedError

  def mean(self):
    raise NotImplementedError

  def max_dev(self):
    raise NotImplementedError

  def l1_dev(self):
    raise NotImplementedError

  def lp_dev(self):
    raise NotImplementedError

  def bound(self, calibration_config):
    raise NotImplementedError

  def check_mean(self, mean):
    self.assertAllEqual(self.mean(), f32(mean))

  def check_max_dev(self, expected_max):
    self.assertAllEqual(self.get_max_of_abs_vals(), f32(expected_max))

  def check_l1_dev(self, l1_dev):
    self.assertAllEqual(self.l1_dev(), f32(l1_dev))

  def check_lp_dev(self, lp_dev, approx=False):
    if approx:
      self.assertAllClose(self.lp_dev(), f32(lp_dev))
    else:
      self.assertAllEqual(self.lp_dev(), f32(lp_dev))

  def test_basics_carefully(self):
    """Checks stats against hand-computed results."""
    config = aqt_test_shared_base.test_stats_config()
    config.share_stats_axes[:] = [1]  # share only on rows

    with self.cached_session():
      rl = 4  # row length
      self.set_stats([2, rl], config)

      # update 1
      # ema_update_count = 1 completely overwrites the priors.
      wgt1 = f32([[2], [0.5]])
      self.update([[4, 0, -3, 0], [-40, 30, 0, 0]], wgt1)
      # We will take a peak inside first.
      # pylint: disable=protected-access
      self.assertAllEqual(self.get_sum_of_ones(), wgt1 * rl)
      self.assertAllEqual(self.get_sum_of_vals(),
                          f32([[4 - 3], [-40 + 30]]) * wgt1)
      self.assertAllEqual(self.get_sum_of_l1_vals(),
                          f32([[4 + 3], [40 + 30]]) * wgt1)
      squre_sum = f32([[4 * 4 + 3 * 3], [40 * 40 + 30 * 30]])
      self.assertAllEqual(self.get_sum_of_lp_vals(), squre_sum * wgt1)
      self.check_mean([[0.25], [-2.5]])
      self.check_l1_dev([[1.75], [17.5]])
      self.check_lp_dev([[2.5], [25]])
      self.check_max_dev([[4], [40]])

      # update 2
      # ema_update_count = 1 completely overwrites the update 1.
      # notice the change in weight and _sum_of_ones, but no change in stats.
      self.update([[4, 0, -3, 0], [-40, 30, 0, 0]], [[1], [0.5]])
      self.assertAllEqual(self.get_sum_of_ones(), f32([[1], [0.5]]) * rl)
      self.check_mean([[0.25], [-2.5]])
      self.check_l1_dev([[1.75], [17.5]])
      self.check_lp_dev([[2.5], [25]])
      self.check_max_dev([[4], [40]])

      # update 3
      # ema = 2 will make update 2 and 3 have equal weight.
      # This is the source of the '/ 2'.
      self.set_ema_update_count(2)
      self.update(f32([[10, 10, 0, 0], [50, 0, 0, 0]]), f32([[3], [0.5]]))
      # Functions could make them one-liners and improve readability.
      self.assertAllEqual(self.get_sum_of_ones(),
                          f32([[1 + 3], [0.5 + 0.5]]) / 2 * rl)
      self.assertAllEqual(self.get_sum_of_vals(),
                          f32([[1 * 1 + 20 * 3], [-10 * 0.5 + 50 * 0.5]]) / 2)
      self.assertAllEqual(self.get_sum_of_l1_vals(),
                          f32([[7 * 1 + 20 * 3], [70 * 0.5 + 50 * 0.5]]) / 2)
      self.assertAllEqual(
          self.get_sum_of_lp_vals(),
          f32([[25 * 1 + 200 * 3], [2500 * 0.5 + 2500 * 0.5]]) / 2)
      self.check_mean([[61 / 16], [5]])
      self.check_l1_dev([[67 / 16], [15]])
      self.check_lp_dev([[25 / 4], [25]])
      self.check_max_dev([[4 / 2 + 10 / 2], [50 / 2 + 40 / 2]])

  def test_filter_zeros_and_bound(self):
    """Makes sure that zero filter skips over zeros for statistics."""
    config = aqt_test_shared_base.test_stats_config()
    config.filter_zeros = True
    calibration_config = aqt_config.CalibrationConfig(
        const_bound_coeff=1,
        l1_dev_coeff=10,
        lp_dev_coeff=100,
        max_dev_coeff=1000,
    )

    with self.cached_session():
      self.set_stats([1, 4], config)
      self.update([[-3, 3, 15, 0]], [[42]])
      self.check_mean([[5]])
      # a non-trivial case where l1_dev, lp_dev and max_dev of different
      # values to test each of them matches expected.
      self.check_l1_dev([[7]])
      self.check_lp_dev([[9]])
      self.check_max_dev([[15]])
      # bound
      self.assertAllEqual(
          self.bound(calibration_config),
          f32([[1 + 10 * 7 + 100 * 9 + 1000 * 15]])
          )

  @parameterized.named_parameters(
      dict(inp=[[0, 0, 0, 0, 0]], w=[[42]], mean=0, l1=0, lp=0, mx=0, bd=10,
           testcase_name="zero_inputs"),
      dict(inp=[[1, 1, 1, -1, 0]], w=[[0]], mean=0, l1=0, lp=0, mx=0, bd=10,
           testcase_name="zero_weights"),
      dict(inp=[[1, 1, 1, -1, 0]], w=[[2]], mean=.5, l1=1, lp=1, mx=1, bd=117,
           testcase_name="non_zeros"),
      )
  def test_filter_zeros_and_safe_divide(self, inp, w, mean, l1, lp, mx, bd):
    """Makes sure that zero filter does not break statistics calculation."""
    config = aqt_test_shared_base.test_stats_config()
    config.filter_zeros = True
    config.safe_divide = True
    calibration_config = aqt_config.CalibrationConfig(
        const_bound_coeff=10, l1_dev_coeff=3, lp_dev_coeff=4, max_dev_coeff=100)

    with self.cached_session():
      self.set_stats([1, 5], config)
      self.update(inp, w)
      self.check_mean([[mean]])
      self.check_l1_dev([[l1]])
      self.check_lp_dev([[lp]])
      self.check_max_dev([[mx]])
      # bound
      self.assertAllEqual(
          self.bound(calibration_config), f32([[bd]]))

  def test_p_norm(self):
    config = aqt_test_shared_base.test_stats_config()
    config.lp_order = 20  # close to max norm

    with self.cached_session():
      self.set_stats([1, 5], config)
      self.update([[1, -2, 3, -40, 0]], [[42]])
      self.check_lp_dev([[36.907234]])  # close to 40

  def test_max_skips_zero_weight(self):
    """Maximum calculation should elide zero-weight rows."""
    config = aqt_test_shared_base.test_stats_config()

    with self.cached_session():
      self.set_stats([1, 1], config)
      self.update([[100], [200]], [[1], [0]])
      self.check_max_dev([[100]])

  def test_odd_norm(self):
    """Validates that odd norms use absolute values."""
    config = aqt_test_shared_base.test_stats_config()
    config.lp_order = 3
    config.ema_update_count = 1

    with self.cached_session():
      acts = [1, -2, 3, -1, 0]
      self.set_stats([1, len(acts)], config)
      self.update([acts], [[1]])
      expected = (sum(abs(x)**3 for x in acts) / len(acts))**(1 / 3)
      self.check_lp_dev([[expected]], approx=True)

  # here we test some properties for a change, any dyadic params should work
  @parameterized.named_parameters(
      dict(op=10, mp=20, l1p=30, lpp=40, mxp=50, ema=8, testcase_name="1"),
      dict(op=10, mp=20, l1p=30, lpp=40, mxp=50, ema=1, testcase_name="2"),
      dict(op=40, mp=30, l1p=20, lpp=10, mxp=100, ema=8, testcase_name="3"),
  )
  def test_prior_and_ema(self, op, mp, l1p, lpp, mxp, ema):
    """Validates that the priors are used for the initial values of variables."""
    config = aqt_test_shared_base.test_stats_config()
    config.update_count_prior = op
    config.mean_prior = mp
    config.l1_dev_prior = l1p
    config.lp_dev_prior = lpp
    config.max_dev_prior = mxp
    config.ema_update_count = ema

    with self.cached_session():
      self.set_stats([1, 1], config)
      # Testing prior first
      self.check_mean([[config.mean_prior]])
      self.check_l1_dev([[config.l1_dev_prior]])
      self.check_lp_dev([[config.lp_dev_prior]])
      self.check_max_dev([[config.max_dev_prior]])
      # Testing decay of all stats given '0' update.
      r = 1 - 1 / (config.ema_update_count)
      self.update([[0]], [[config.update_count_prior]])
      self.check_mean([[config.mean_prior * r]])
      self.check_l1_dev([[config.l1_dev_prior * r]])
      self.check_lp_dev([[config.lp_dev_prior * r**(1. / config.lp_order)]])
      self.check_max_dev([[config.max_dev_prior * r]])

  def test_no_axis_share(self):
    """Tests the case where no axes have shared stats."""
    config = aqt_test_shared_base.test_stats_config()
    # We have only EMA averaging.
    config.share_stats_axes[:] = []

    with self.cached_session():
      self.set_stats([2, 3], config)
      self.update([[1, 2, -3], [4, 5, -6]], [[1, 1, 1], [1, 1, 1]])
      self.check_mean([[1, 2, -3], [4, 5, -6]])
      self.check_l1_dev([[1, 2, 3], [4, 5, 6]])
      self.check_lp_dev([[1, 2, 3], [4, 5, 6]], approx=True)

  # We average across both axis of 2-dim Tensor
  def test_axis_share_both(self):
    config = aqt_test_shared_base.test_stats_config()
    config.share_stats_axes[:] = [0, 1]

    with self.cached_session():
      self.set_stats([2, 2], config)
      self.update([[1, 2], [0, -2]], [[1, 1], [1, 1]])
      self.check_mean([[0.25]])
      self.check_l1_dev([[1.25]])
      self.check_lp_dev([[1.5]], approx=True)

  def test_cross_tpu(self):
    if not flags.FLAGS.use_tpu:
      return
    return  # TODO(lew): implement.
    # config = aqt_test_shared_base.test_stats_config()
    # config.tpu_cross_replica_sum = True
    # config.share_stats_axes[:] = []
    # with self.cached_session():
    #   self.set_stats([2, 3], config)
    #   self.update([[1, 2, -3], [4, 5, -6]], [[1, 1, 1], [1, 1, 1]])
    #   self.check_mean([[1, 2, -3], [4, 5, -6]])
    #   self.check_l1_dev([[1, 2, 3], [4, 5, 6]])
    #   self.check_lp_dev([[1, 2, 3], [4, 5, 6]], approx=True)
    #   # TODO(lew): How to create multi-replica Stats?



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AqtTensorQuantizer test base to be shared between TF and JAX."""

import copy
import itertools

from absl import logging
from absl.testing import parameterized
from aqt.common import aqt_common
from aqt.common import aqt_config
import numpy as np
import tensorflow.compat.v1 as tf


def f32(x):
  """Cast input array to f32."""
  return np.array(x, dtype=np.float32)


def make_quant_cases():
  for const, freeze in itertools.product([False, True], repeat=2):
    name = f"const_{const}_freeze_{freeze}"
    yield dict(
        const_calibration=const,
        freeze_scale_at_begin=freeze,
        testcase_name=name)


class AqtTensorQuantizerTest(tf.test.TestCase, parameterized.TestCase):
  """Base class for testing AqtTensorQuantizer.

  As TF and JAX AQTp have the same API and logic for AqtTensorQuantizer class,
  all testing cases are put in this base class and are shared between TF and JAX
  through inheritance. On TF and JAX AQTp sides, only few methods which are not
  framework agnostic should be implemented.
  """

  def make_tensor_quantizer(self, data_shape, config, name="tq"):
    raise NotImplementedError

  def update_quantizer(self, quant, sample, weight, event_count):
    raise NotImplementedError

  def to_quant(self, quant, x, train=True):
    raise NotImplementedError

  def get_quant_scale(self, quant, train=True):
    raise NotImplementedError

  def init(self):
    raise NotImplementedError

  def get_scale(self, quant):
    raise NotImplementedError

  def get_clip_range(self, quant):
    raise NotImplementedError

  def get_last_update(self, quant):
    raise NotImplementedError

  def get_quantized_variable(self, quant):
    raise NotImplementedError

  def quantize(self, x, quant, train):
    scale, inv_scale = self.get_quant_scale(quant, train)
    x = scale * x
    ix = self.to_quant(quant, x)
    return inv_scale * ix

  def test_validates_shape_config(self):
    """Checks that initializer validates the shape against config."""
    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=[1],
        tpu_cross_replica_sum=False,
    )
    config = aqt_config.AqtTensorConfig(
        quant_config=aqt_config.IntQuantConfig(bits=8),
        calibration_config=aqt_config.CalibrationConfig(const_bound_coeff=1),
        freeze_scale_at_begin=True)
    config = aqt_config.AqtScheduleConfig(sc, [config])

    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "share_stats_axes .* contain unknown"):
      self.make_tensor_quantizer(data_shape=[None, 3], config=config)

    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "quantized variable with unknown"):
      config.use_quantized_variable = True
      self.make_tensor_quantizer(data_shape=[3, None], config=config)

  def test_validates_shape_update(self):
    """Check that update() validates input tensor shapes."""
    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=[0],
        tpu_cross_replica_sum=False,
    )
    config = aqt_config.AqtTensorConfig(
        quant_config=aqt_config.IntQuantConfig(bits=8),
        calibration_config=aqt_config.CalibrationConfig(const_bound_coeff=1),
        freeze_scale_at_begin=True)
    config = aqt_config.AqtScheduleConfig(sc, [config])

    quant = self.make_tensor_quantizer(data_shape=[None, 3], config=config)

    self.init()

    for good_shape in [(1, 3), (3, 3)]:
      x = np.ones(good_shape, dtype=np.float32)
      self.update_quantizer(quant, x, None, 0)

    for bad_shape in [(3,), (1, 3, 2)]:
      with self.assertRaisesRegex(ValueError, "shape .* compatible with"):
        x = np.ones(bad_shape, dtype=np.float32)
        self.update_quantizer(quant, x, None, 0)

  def test_assert_min_event(self):
    """Validates that event_count supplied cannot be int64.min."""
    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=[0, 1],
        tpu_cross_replica_sum=False,
    )
    config = aqt_config.AqtTensorConfig(
        quant_config=aqt_config.IntQuantConfig(bits=8),
        calibration_config=aqt_config.CalibrationConfig(const_bound_coeff=1),
        freeze_scale_at_begin=True)
    config = aqt_config.AqtScheduleConfig(sc, [config])

    x = np.ones((4, 4), dtype=np.float32)
    quant = self.make_tensor_quantizer(data_shape=x.shape, config=config)
    event_count = np.iinfo(np.int64).min
    with self.assertRaisesRegex(Exception,
                                "event_count cannot be"):
      self.update_quantizer(quant, x, None, event_count)

  def test_single_quant_simple(self):
    """Compares quantization to hand-computed example."""
    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=[0, 1],
        tpu_cross_replica_sum=False,
    )
    config = aqt_config.AqtTensorConfig(
        quant_config=aqt_config.IntQuantConfig(bits=3),
        calibration_config=aqt_config.CalibrationConfig(const_bound_coeff=7),
        freeze_scale_at_begin=True,
    )
    # representable values: -6, -4, -2, 0, 2, 4, 6

    x = f32([
        [0.99, 1.01, 1.99, 2.01],  #
        [2.99, 3.01, 3.99, 4.01],  #
        [4.99, 5.01, 5.99, 6.01],  #
        [6.99, 7.01, 7.99, 8.01],  #
        [-0.99, -1.01, -1.99, -2.01],  #
        [-2.99, -3.01, -3.99, -4.01],  #
        [-4.99, -5.01, -5.99, -6.01],  #
        [-6.99, -7.01, -7.99, -8.01],  #
    ])
    expected_output = f32([
        [0.00, 2.00, 2.00, 2.00],  #
        [2.00, 4.00, 4.00, 4.00],  #
        [4.00, 6.00, 6.00, 6.00],  #
        [6.00, 6.00, 6.00, 6.00],  #
        [-0.00, -2.00, -2.00, -2.00],  #
        [-2.00, -4.00, -4.00, -4.00],  #
        [-4.00, -6.00, -6.00, -6.00],  #
        [-6.00, -6.00, -6.00, -6.00],  #
    ])

    config = aqt_config.AqtScheduleConfig(sc, [config])
    quant = self.make_tensor_quantizer(data_shape=[8, 4], config=config)

    self.init()
    event_count = np.array(0, dtype=np.int64)
    self.update_quantizer(quant, x, np.full((1, 1), 1, dtype=np.float32),
                          event_count)
    qx = self.quantize(x, quant, True)
    self.assertAllEqual(self.get_scale(quant), np.full((1, 1), 0.5,
                                                       dtype=np.float32))
    self.assertAllEqual(qx, expected_output)

  @parameterized.named_parameters(make_quant_cases())
  def test_quant(self, const_calibration, freeze_scale_at_begin):
    """Tests basic TensorQuantizer behavior between update and to_quant."""
    bits = 8
    x_bound = 16.0

    calibration_config = aqt_config.CalibrationConfig(lp_dev_coeff=3)
    if const_calibration:
      calibration_config = aqt_config.CalibrationConfig(
          const_bound_coeff=x_bound)

    sc = aqt_config.StatsConfig(
        ema_update_count=1, share_stats_axes=[1], tpu_cross_replica_sum=False)
    config = aqt_config.AqtTensorConfig(
        quant_config=aqt_config.IntQuantConfig(bits),
        calibration_config=calibration_config,
        freeze_scale_at_begin=freeze_scale_at_begin,
        begin_at_event=10,
        end_at_event=20,
    )
    config = aqt_config.AqtScheduleConfig(sc, [config])

    data_shape = [128, 256]
    has_quantized = False
    quant = self.make_tensor_quantizer(data_shape, config, "tq")
    self.init()
    last_scale = self.get_scale(quant)
    logging.info("scale l1 before loop: %s", np.abs(last_scale).sum())
    for (new_event_count, should_quantize) in [(9, False), (10, True),
                                               (19, True), (20, False)]:
      logging.info("loop: %s", (new_event_count, should_quantize))
      x = f32(np.random.uniform(low=-x_bound, high=x_bound, size=data_shape))
      bucket_count = 2.0**bits - 1
      bucket_size = 2 * x_bound / bucket_count
      new_event_count = np.array(new_event_count, dtype=np.int64)
      self.update_quantizer(quant,
                            x,
                            np.full((data_shape[0], 1), 1, dtype=np.float32),
                            new_event_count)
      logging.info("  scale l1: %s", np.abs(self.get_scale(quant)).sum())

      self.assertAllEqual(self.get_last_update(quant), new_event_count)

      scale, inv_scale = self.get_quant_scale(quant)
      x_scaled = scale * x

      ix = self.to_quant(quant, x_scaled)
      has_quantized |= should_quantize

      self.assertAllEqual(
          self.get_scale(quant).shape,
          self.get_clip_range(quant).shape)

      if should_quantize:
        # all quantized values are integers in a proper range
        self.assertAllEqual(ix, np.round(ix))
        self.assertAllLessEqual(np.abs(ix), 2**(bits - 1) - 1)
        bound = self.get_clip_range(quant)
        if const_calibration:
          self.assertAllEqual(bound, np.ones_like(bound) * x_bound)
        self.assertAllClose(bound, 127.5 / self.get_scale(quant))
      else:
        self.assertAllEqual(ix, x)
        bound = self.get_clip_range(quant)
        self.assertAllEqual(bound, np.zeros_like(bound))

      if const_calibration and has_quantized:
        self.assertAllEqual(self.get_scale(quant),
                            np.full((data_shape[0], 1), 127.5 / x_bound,
                                    dtype=np.float32))
      if not const_calibration:
        # does freeze_scale_at_begin has a proper effect
        curr_scale = self.get_scale(quant)
        scale_not_changed = (curr_scale == last_scale).all()
        frozen = should_quantize and freeze_scale_at_begin
        not_quantized = not should_quantize
        first_frozen_event = new_event_count == 10
        frozen &= not first_frozen_event  # First frozen event requires update.
        logging.info(
            "  scale_not_changed=%s,not_quantized=%s,frozen=%s,"
            "first_frozen_event=%s", scale_not_changed, not_quantized, frozen,
            first_frozen_event)
        self.assertEqual(scale_not_changed, not_quantized or frozen)
        last_scale = curr_scale

      qx = inv_scale * ix
      quant_noise = np.abs(qx - np.clip(x, -x_bound, x_bound))
      if should_quantize:
        if const_calibration:
          # Checking quant_noise is the core of this test.
          # its properties should hold in all configurations.
          # 1. max error should be half the bucket.
          self.assertAllLessEqual(
              np.max(quant_noise), bucket_size / 2)
          # 2. error size should be half of max error
          self.assertAllClose(
              np.mean(quant_noise),
              bucket_size / 4,
              atol=bucket_size / 200)
      else:
        self.assertAllEqual(quant_noise, np.zeros_like(quant_noise))

  def test_update_saves_variable(self):
    """Makes sure that update saves to quantized_variable."""
    iqc = aqt_config.IntQuantConfig(bits=8, preserve_zero=True)
    clip_bound = aqt_common.get_clip_bound(iqc)
    # Use a half of clip_bound for const_bound_coeff, which results in
    # scale=2.0. It will help to easily compute expected quantized values and
    # check if they are equal to actual quantized variable in this test.
    cc = aqt_config.CalibrationConfig(const_bound_coeff=clip_bound * 0.5)
    sc = aqt_config.StatsConfig(
        ema_update_count=1, share_stats_axes=[], tpu_cross_replica_sum=False)
    yes_q_config = aqt_config.AqtTensorConfig(
        freeze_scale_at_begin=True,
        quant_config=iqc,
        calibration_config=cc,
        end_at_event=1)

    # Create a quantizer with the first config quantized, the second not.
    iqc = aqt_config.IntQuantConfig(bits=9, preserve_zero=True)
    clip_bound = aqt_common.get_clip_bound(iqc)

    no_q_config = copy.deepcopy(yes_q_config)
    yes_q_config.end_at_event = 1
    no_q_config.quant_config = iqc
    no_q_config.calibration_config.const_bound_coeff = clip_bound
    no_q_config.begin_at_event = 1
    no_q_config.end_at_event = 2
    tensor_configs = [yes_q_config, no_q_config]
    config = aqt_config.AqtScheduleConfig(sc, tensor_configs, True)

    data_shape = (4, 2)
    quantizer = self.make_tensor_quantizer(data_shape=data_shape, config=config,
                                           name="quantizer")

    def update(sample, event_count):
      event_count = np.array(event_count, dtype=np.int64)
      self.update_quantizer(quantizer, sample, None, event_count)

    rng = np.random.default_rng(1234)
    x = rng.integers(-10, 10, size=data_shape, dtype=np.int64)
    x = np.array(x, dtype=np.float32)
    # Since all values in x are integers between -10 and 10, their quantized
    # values are just the result of scaling (scale=2.0), without clipping
    # (clip_bound [-127.5, 127.5]) and rounding.
    x_quantized = 2.0 * x

    self.init()

    z = np.zeros_like(self.get_quantized_variable(quantizer))
    self.assertAllEqual(self.get_quantized_variable(quantizer), z)

    # First config with 8 bits quantization active, variable should be updated.
    update(x, 0)
    self.assertAllEqual(self.get_quantized_variable(quantizer), x_quantized)

    # Second config with 9 bits quantization active, variable should not be
    # updated, since it is not compatible with int8.
    update(x * 2, 1)
    self.assertAllEqual(self.get_quantized_variable(quantizer), x_quantized)

    # No config, variable should not be updated.
    update(x * 2, 2)
    self.assertAllEqual(self.get_quantized_variable(quantizer), x_quantized)

  def test_none_weights(self):
    """Ensures semantics of None weights equal those of weights=1."""
    iqc = aqt_config.IntQuantConfig(bits=8, preserve_zero=True)
    cc = aqt_config.CalibrationConfig(l1_dev_coeff=2)
    sc = aqt_config.StatsConfig(
        ema_update_count=10,
        share_stats_axes=[0, 1],
        tpu_cross_replica_sum=False)
    tc = aqt_config.AqtTensorConfig(
        quant_config=iqc,
        calibration_config=cc,
        freeze_scale_at_begin=False)
    config = aqt_config.AqtScheduleConfig(sc, [tc], use_quantized_variable=True)

    rng = np.random.default_rng(1234)
    x = rng.normal(size=(3, 4)).astype(np.float32)
    q1 = self.make_tensor_quantizer(data_shape=x.shape, config=config,
                                    name="q1")
    q2 = self.make_tensor_quantizer(data_shape=x.shape, config=config,
                                    name="q2")
    weight = np.full(x.shape, 1.0, dtype=np.float32)

    self.init()

    ec = np.array(0, dtype=np.int64)
    self.update_quantizer(q1, x, None, ec)
    self.update_quantizer(q2, x, weight, ec)

    self.assertAllEqual(self.get_quantized_variable(q1),
                        self.get_quantized_variable(q2))

  def test_float_config_not_save_quantized_var(self):
    """Ensures FloatConfig does not save quantized variables."""
    float_tc = aqt_config.AqtTensorConfig(
        quant_config=aqt_config.FloatConfig(),
        calibration_config=aqt_config.CalibrationConfig(),
        freeze_scale_at_begin=True,
        begin_at_event=0,
        end_at_event=1)
    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=[0, 1],
        tpu_cross_replica_sum=False)
    use_quantized_var = True
    config = aqt_config.AqtScheduleConfig(sc, [float_tc], use_quantized_var)

    rng = np.random.default_rng(1234)
    x = rng.normal(size=(3, 4)).astype(np.float32)
    quantizer = self.make_tensor_quantizer(data_shape=x.shape, config=config)

    self.init()

    event_count = np.array(0, dtype=np.int64)
    self.update_quantizer(quantizer, x, None, event_count)

    actual = self.get_quantized_variable(quantizer)
    expected = np.zeros_like(actual)
    self.assertAllEqual(actual, expected)

  def test_float_config_not_quantized(self):
    """Ensures an input is not quantized when it comes with FloatConfig."""
    iqc = aqt_config.IntQuantConfig(bits=3, preserve_zero=True)
    cc = aqt_config.CalibrationConfig(const_bound_coeff=7)
    int_tc = aqt_config.AqtTensorConfig(
        quant_config=iqc,
        calibration_config=cc,
        freeze_scale_at_begin=True,
        begin_at_event=1,
        end_at_event=2)
    float_tc = copy.deepcopy(int_tc)
    float_tc.quant_config = aqt_config.FloatConfig()
    float_tc.begin_at_event = 0
    float_tc.end_at_event = 1

    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=[0, 1],
        tpu_cross_replica_sum=False)

    tensor_configs = [float_tc, int_tc]
    config = aqt_config.AqtScheduleConfig(sc, tensor_configs, True)

    x = f32([
        [0.99, 1.01, 1.99, 2.01],  #
        [2.99, 3.01, 3.99, 4.01],  #
        [4.99, 5.01, 5.99, 6.01],  #
        [6.99, 7.01, 7.99, 8.01],  #
        [-0.99, -1.01, -1.99, -2.01],  #
        [-2.99, -3.01, -3.99, -4.01],  #
        [-4.99, -5.01, -5.99, -6.01],  #
        [-6.99, -7.01, -7.99, -8.01],  #
    ])
    x_quantized = f32([
        [0.00, 2.00, 2.00, 2.00],  #
        [2.00, 4.00, 4.00, 4.00],  #
        [4.00, 6.00, 6.00, 6.00],  #
        [6.00, 6.00, 6.00, 6.00],  #
        [-0.00, -2.00, -2.00, -2.00],  #
        [-2.00, -4.00, -4.00, -4.00],  #
        [-4.00, -6.00, -6.00, -6.00],  #
        [-6.00, -6.00, -6.00, -6.00],  #
    ])

    quantizer = self.make_tensor_quantizer(data_shape=x.shape, config=config,
                                           name="quantizer")

    self.init()

    def update_and_quantize(event_count):
      event_count = np.array(event_count, dtype=np.int64)
      self.update_quantizer(quantizer, x, None, event_count)
      qx = self.quantize(x, quantizer, True)
      return qx

    # At event_count=0, the input should not be quantized since the active
    # config is FloatConfig.
    qx = update_and_quantize(0)
    self.assertAllEqual(x, qx)

    # At event_count=1, the input should be quantized since the active config is
    # IntQuantConfig.
    qx = update_and_quantize(1)
    self.assertAllEqual(x_quantized, qx)

  @parameterized.parameters([{
      "use_quantized_variable": True
  }, {
      "use_quantized_variable": False
  }])
  def test_inference_config_index(self, use_quantized_variable):
    """Validate that the inference config index is used."""

    # Create a schedule where quantization behavior at times 0 and 1
    # differ. Then make sure that a dynamically switched inference
    # (with the inference index set to None) matches the appropriate
    # value for a statistically switched inference quantizer.

    iqc = aqt_config.IntQuantConfig(bits=8, preserve_zero=True)
    cc = aqt_config.CalibrationConfig(const_bound_coeff=2)
    int_tc = aqt_config.AqtTensorConfig(
        quant_config=iqc,
        calibration_config=cc,
        freeze_scale_at_begin=True,
        begin_at_event=1,
        end_at_event=2)
    float_tc = copy.deepcopy(int_tc)
    float_tc.quant_config = aqt_config.FloatConfig()
    float_tc.begin_at_event = 0
    float_tc.end_at_event = 1
    tensor_configs = [float_tc, int_tc]
    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=[0, 1],
        tpu_cross_replica_sum=False)
    config_no_inference_index = aqt_config.AqtScheduleConfig(
        sc, tensor_configs, use_quantized_variable)
    config_inference0 = copy.deepcopy(config_no_inference_index)
    config_inference0.inference_config_index = 0
    config_inference1 = copy.deepcopy(config_no_inference_index)
    config_inference1.inference_config_index = 1

    data_shape = (3, 4)
    q_index_none = self.make_tensor_quantizer(
        data_shape, config=config_no_inference_index)
    q_index_0 = self.make_tensor_quantizer(
        data_shape, config=config_inference0, name="0")
    q_index_1 = self.make_tensor_quantizer(
        data_shape, config=config_inference1, name="1")

    self.init()

    # Note that because we only use constant coeff, updates don't
    # affect true scale values; only the config choice does.
    def update(event_count, quantizer):
      event_count = np.array(event_count, dtype=np.int64)
      x = np.ones(data_shape).astype(np.float32)
      self.update_quantizer(quantizer, x, None, event_count)

    # Update the indexed quantizers at least once to initialize.
    update(0, q_index_0)
    update(0, q_index_1)

    rng = np.random.default_rng(1234)

    def check_quant(q_index, train):
      """Checks q_index has the same behavior as q_index_none."""
      x = rng.normal(size=(3, 4)).astype(np.float32)
      scale_unindexed, inv_scale_unindexed = self.get_quant_scale(
          q_index_none, train=train)
      x_unindexed = scale_unindexed * x
      unindexed = self.to_quant(q_index_none, x_unindexed, train=train)
      scale_indexed, inv_scale_indexed = self.get_quant_scale(
          q_index, train=train)
      x_indexed = scale_indexed * x
      indexed = self.to_quant(q_index, x_indexed, train=train)

      self.assertAllEqual(unindexed, indexed)
      self.assertAllEqual(inv_scale_unindexed, inv_scale_indexed)

    update(0, q_index_none)
    check_quant(q_index_0, train=True)
    check_quant(q_index_0, train=False)

    # Even if we update the statically indexed one its behavior is the same at
    # inference time.
    update(1, q_index_0)
    check_quant(q_index_0, train=False)

    # Now update our dynamically indexed quantizer to time 1.
    update(1, q_index_none)
    check_quant(q_index_0, train=True)  # recall this is still at time 1.

    update(1, q_index_1)
    check_quant(q_index_1, train=True)
    check_quant(q_index_1, train=False)

  @parameterized.parameters([{
      "shared_stats_axes": [0],
  }, {
      "shared_stats_axes": [0, 1],
  }, {
      "shared_stats_axes": [0, 1, 2],
  }])
  def test_scale_and_inv_scale(self, shared_stats_axes):
    """Validates scale * inv_scale close to an identity tensor."""
    iqc = aqt_config.IntQuantConfig(bits=8, preserve_zero=True)
    cc = aqt_config.CalibrationConfig(l1_dev_coeff=2)
    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=shared_stats_axes,
        tpu_cross_replica_sum=False)
    tc = aqt_config.AqtTensorConfig(
        quant_config=iqc,
        calibration_config=cc,
        freeze_scale_at_begin=False)
    config = aqt_config.AqtScheduleConfig(sc, [tc], use_quantized_variable=True)

    rng = np.random.default_rng(1234)
    x = rng.normal(size=(2, 3, 4)).astype(np.float32)
    quantizer = self.make_tensor_quantizer(x.shape, config)

    self.init()
    self.update_quantizer(quantizer, x, None, 0)

    scale, inv_scale = self.get_quant_scale(quantizer, train=True)

    actual = scale * inv_scale
    expected = np.ones_like(actual)

    self.assertAllClose(actual, expected)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import concurrent.futures
import math
import re

from absl import flags
from absl.testing import absltest
from absl.testing import flagsaver
from absl.testing import parameterized
from aqt.jax_legacy.jax.wmt_mlperf import train
from aqt.jax_legacy.jax.wmt_mlperf import training_hparams
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental import minimal_model_8bit_weights_and_auto_acts
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental import minimal_model_bfloat16
from aqt.jax_legacy.jax.wmt_mlperf.training_hparams_generator_lib import BaseConfigSize
from aqt.jax_legacy.jax.wmt_mlperf.training_hparams_generator_lib import create_training_hparams_from_flags
from aqt.jax_legacy.utils import hparams_utils as os_hparams_utils
import jax
from jax import numpy as jnp
import numpy as np
import tensorflow.compat.v2 as tf

FLAGS = flags.FLAGS


def create_1x1():
  return np.ones((1, 1), dtype=np.int64)


def create_mock_data():
  """Create the absolute smallest dataset that can be supplied to the training loop."""

  # Note that not all these keys are necessary for each kind of dataset
  # (train, eval, predict),
  # but it doesn't hurt to have them.
  data = {
      'inputs': create_1x1(),
      'inputs_position': create_1x1(),
      'inputs_segmentation': create_1x1(),
      'targets': create_1x1(),
      'targets_position': create_1x1(),
      'targets_segmentation': create_1x1(),
  }
  dataset = tf.data.Dataset.from_tensor_slices(data).batch(1)
  return dataset


# The training loop requires a decoder for its target language (ie, word
# index->human-readable string) to produce human-readable output during
# evaluation, so we mock the decoder.
class MockEncoder():
  """Mocks the tensorflow_text.SentencepieceTokenizer used in input_pipeline.py."""

  def vocab_size(self):
    # Not '1' so that 'low_confidence' calculation in train.py
    # doesn't divide-by-zero.
    return 2

  def detokenize(self, _):
    return tf.constant('<mock>')


# TODO(wanglisa): Once the remaining transformer kwargs are moved to flags,
# these test cases can all be specified in terms of flags (and thus we gain
# additional test coverage of the flag-processing functions in
# training_hparams.py).
class TrainTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.datasets = train.Datasets(
        train_ds=create_mock_data(),
        eval_ds_dict={'mock_data': create_mock_data()},
        train_eval_ds=create_mock_data(),
        predict_ds_dict={'mock_data': create_mock_data()},
        encoder=MockEncoder())

  @parameterized.named_parameters(
      dict(
          testcase_name='minimal_bfloat16',
          hparams_config_filename=minimal_model_bfloat16),)
  def test_train_with_config_file(self, hparams_config_filename):
    with flagsaver.flagsaver(
        num_eval_steps=1,
        eval_batch_size=1,
        max_target_length=1,
        eval_dataset_name='mock_data',
        max_eval_target_length=1,
        max_predict_length=1,
        model_dir=FLAGS.test_tmpdir,
    ):
      hparams = os_hparams_utils.load_dataclass_from_config_dict(
          training_hparams.TrainingHParams,
          hparams_config_filename.get_config())
      with concurrent.futures.ThreadPoolExecutor(max_workers=1) as io_executor:
        train.run_training(
            datasets=self.datasets,
            hparams=hparams,
            io_executor=io_executor,
        )

  def test_checkpointing(self):
    hparams = os_hparams_utils.load_dataclass_from_config_dict(
        training_hparams.TrainingHParams,
        minimal_model_8bit_weights_and_auto_acts.get_config())
    training_state_initial = train.TrainingState.initialize(
        encoder=self.datasets.encoder, hparams=hparams)

    ckpt_dir = FLAGS.test_tmpdir + '/ckpt_dir'
    tf.io.gfile.makedirs(ckpt_dir)
    # Delete existing files in the temp directory to clear out old checkpoints.
    for dir_name, _, file_names in tf.io.gfile.walk(ckpt_dir):
      for filename in file_names:
        file_path = tf.io.gfile.join(dir_name, filename)
        tf.io.gfile.remove(file_path)

    training_state, _ = train.run_train_step(
        training_state=training_state_initial,
        batch=next(iter(self.datasets.train_ds)),
        step=0,
        hparams=hparams)

    self.assertFalse(train.does_checkpoint_exist(ckpt_dir))
    training_state.save_checkpoint(model_dir=ckpt_dir, step=0)
    self.assertTrue(train.does_checkpoint_exist(ckpt_dir))

    self.assertNotEqual(training_state.flax_state,
                        training_state_initial.flax_state)
    with np.testing.assert_raises(AssertionError):
      np.testing.assert_array_equal(training_state.dropout_rngs,
                                    training_state_initial.dropout_rngs)
    training_state_restored = training_state_initial.restore_checkpoint(
        model_dir=ckpt_dir)

    leaf_equality_tree = jax.tree_map(lambda x, y: jnp.all(x == y),
                                           training_state.flax_state,
                                           training_state_restored.flax_state)
    self.assertTrue(
        all(jax.tree_leaves(leaf_equality_tree)),
        'Training state was altered during restoration.')

    np.testing.assert_array_equal(training_state.dropout_rngs,
                                  training_state_restored.dropout_rngs)


class LearningRateSchedulerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.hparams = training_hparams.LearningRateSchedulerHParams(
        factors='',
        base_learning_rate=2.0,
        warmup_steps=0,
        decay_factor=None,
        steps_per_decay=None,
        steps_per_cycle=None)

  def test_constant(self):
    hparams = self.hparams
    hparams.factors = 'constant'
    schedule = train.create_learning_rate_scheduler(hparams)
    self.assertEqual(schedule(0), 2.0)
    self.assertEqual(schedule(100), 2.0)

  def test_linear_warmup(self):
    hparams = self.hparams
    hparams.factors = 'constant*linear_warmup'
    hparams.warmup_steps = 10
    schedule = train.create_learning_rate_scheduler(hparams)
    self.assertEqual(schedule(0), 0.0)
    self.assertEqual(schedule(5), 2.0 / 2)
    self.assertEqual(schedule(10), 2.0)
    self.assertEqual(schedule(100), 2.0)

  def test_rsqrt_decay(self):
    hparams = self.hparams
    hparams.factors = 'constant*rsqrt_decay'
    hparams.warmup_steps = 9
    hparams.base_learning_rate = 1.0
    schedule = train.create_learning_rate_scheduler(hparams)
    self.assertEqual(schedule(0), 1 / math.sqrt(9))
    self.assertEqual(schedule(9), 1 / math.sqrt(9))
    self.assertEqual(schedule(100), 1 / math.sqrt(100))

  def test_mlperf_schedule(self):
    # Parameters modified slightly from the defaults to avoid dealing with
    # floating point error considerations when taking square roots.
    hparams = training_hparams.LearningRateSchedulerHParams(
        factors='constant * linear_warmup * rsqrt_decay',
        base_learning_rate=0.5,
        warmup_steps=100,
        decay_factor=0.5,
        steps_per_decay=20000,
        steps_per_cycle=100000)
    schedule = train.create_learning_rate_scheduler(hparams)
    self.assertEqual(schedule(0), 0.0)
    self.assertEqual(schedule(50), 0.5 / 2 * 1 / math.sqrt(100))
    self.assertEqual(schedule(100), 0.5 * 1 / math.sqrt(100))
    self.assertEqual(schedule(900), 0.5 * 1 / math.sqrt(900))


class BestCheckpointTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.model_dir = FLAGS.test_tmpdir + '/model_dir'
    tf.io.gfile.makedirs(self.model_dir)
    # Delete existing files in the temp directory to clear out old checkpoints.
    for dir_name, _, file_names in tf.io.gfile.walk(self.model_dir):
      for filename in file_names:
        file_path = tf.io.gfile.join(dir_name, filename)
        tf.io.gfile.remove(file_path)

  def save_checkpoint(self, loss):
    with flagsaver.flagsaver(base_config_size=BaseConfigSize.MINIMAL_MODEL):
      hparams = create_training_hparams_from_flags()
    training_state = train.TrainingState.initialize(
        encoder=MockEncoder(), hparams=hparams)
    train.save_best_checkpoint(
        model_dir=self.model_dir, training_state=training_state, loss=loss)

  def get_model_dir_files(self):
    filenames = list(tf.io.gfile.walk(self.model_dir))[0][2]
    return filenames

  def get_checkpoint(self) -> str:
    """Gets the name of a saved checkpoint."""
    files = self.get_model_dir_files()
    # Test there is exactly one saved checkpoint.
    self.assertLen(
        files, 1, f'Expected exactly one checkpoint to be found, got {files}.')

    # Extract the loss associated with the checkpoint.
    checkpoint_match = re.match(r'.*best_checkpoint_eval_loss_(.*)$',
                                str(files[0]))
    self.assertIsNotNone(checkpoint_match)
    self.assertLen(checkpoint_match.groups(), 1)
    return checkpoint_match.groups()[0]

  def test_min_loss_saved(self):
    self.save_checkpoint(2.0)
    self.save_checkpoint(1.0)
    self.save_checkpoint(3.0)
    self.assertEqual(self.get_checkpoint(), '1.0')

  @parameterized.named_parameters(
      dict(testcase_name='nan', bad_value=math.nan),
      dict(testcase_name='inf', bad_value=math.inf),
      dict(testcase_name='-inf', bad_value=-math.inf))
  def test_nonfinite_ignored(self, bad_value):
    self.save_checkpoint(2.0)
    self.save_checkpoint(bad_value)
    self.save_checkpoint(1.0)
    self.save_checkpoint(3.0)
    self.assertEqual(self.get_checkpoint(), '1.0')

  def test_duplicate_loss(self):
    self.save_checkpoint(1.0)
    self.save_checkpoint(2.0)
    self.save_checkpoint(1.0)
    self.save_checkpoint(3.0)
    self.assertEqual(self.get_checkpoint(), '1.0')

  def test_no_checkpoint_saved_if_all_nan(self):
    for _ in range(5):
      self.save_checkpoint(math.nan)
    self.assertEmpty(self.get_model_dir_files())


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared implementation of prediction step for gen_hlo.py and train.py."""
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax.wmt_mlperf import decode
from aqt.jax_legacy.jax.wmt_mlperf import models
import jax.numpy as jnp


def step(inputs, params, cache, state, eos_token, max_decode_len,
         transformer_kwargs, hparams: models.Transformer.HParams,
         dynamic_context: quant_config.DynamicContext):
  """Predict translation with fast decoding beam search on a batch."""
  # If the state already has a cache, remove it in favor the 'cache' parameter
  # passed to this function.
  if 'cache' in state:
    raise ValueError(
        'cache was already specified in the `state` variable '
        'passed to `step`, which we disallow since it is ambiguous with the '
        '`cache` argument passed to `step`.')
  batch_size = inputs.shape[0]
  beam_size = 4

  # Prepare transformer fast-decoder call for beam search:
  # for beam search, we need to set up our decoder model
  # to handle a batch size equal to batch_size * beam_size,
  # where each batch item's data is expanded in-place rather
  # than tiled.
  # i.e. if we denote each batch element subtensor as el[n]:
  # [el0, el1, el2] --> beamsize=2 --> [el0,el0,el1,el1,el2,el2]
  src_padding_mask = decode.flat_batch_beam_expand((inputs > 0)[..., None],
                                                   beam_size)
  tgt_padding_mask = decode.flat_batch_beam_expand(
      jnp.ones((batch_size, 1, 1)), beam_size)
  model = models.Transformer(
      **transformer_kwargs,
      train=False,
      dynamic_context=dynamic_context,
      hparams=hparams,
      should_decode=False,
      dropout_rate=0.0,
      attention_dropout_rate=0.0,
      use_bfloat16=False)
  mutable = ['sparsity']
  new_logits, _ = model.apply({
      'params': params,
      **state
  },
                              inputs,
                              method=model.encode,
                              mutable=mutable)
  encoded_inputs = decode.flat_batch_beam_expand(
      new_logits, beam_size)

  def tokens_ids_to_logits(flat_ids, flat_cache):
    """Token slice to logits from decoder model."""
    # --> [batch * beam, 1, vocab]
    model = models.Transformer(
        **transformer_kwargs,
        train=False,
        dynamic_context=dynamic_context,
        hparams=hparams,
        should_decode=True,
        dropout_rate=0.0,
        attention_dropout_rate=0.0,
        use_bfloat16=False)
    mutable = ['cache']
    flat_logits, new_vars = model.apply(
        {
            'params': params,
            'cache': flat_cache,
            **state
        },
        encoded=encoded_inputs,
        src_padding_mask=src_padding_mask,
        targets=flat_ids,
        tgt_padding_mask=tgt_padding_mask,
        method=model.decode,
        mutable=mutable)
    new_flat_cache = new_vars['cache']

    return flat_logits, new_flat_cache

  # using the above-defined single-step decoder function, run a
  # beam search over possible sequences given input encoding.
  beam_seqs, _ = decode.beam_search(
      inputs,
      cache,
      tokens_ids_to_logits,
      beam_size=beam_size,
      alpha=0.6,
      eos_token=eos_token,
      max_decode_len=max_decode_len)

  # beam search returns [n_batch, n_beam, n_length + 1] with beam dimension
  # sorted in increasing order of log-probability
  # return the highest scoring beam sequence, drop first dummy 0 token.
  return beam_seqs[:, -1, 1:]



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Numerics for fp8."""

from typing import Literal, Optional, TypeAlias
from aqt.jax.v2 import stochastic_rounding
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import numerics
import jax
import jax.numpy as jnp


FP8Dtype: TypeAlias = Literal['e4m3', 'e5m2']
fp8_map = {'e4m3': jnp.float8_e4m3fn, 'e5m2': jnp.float8_e5m2}


@utils.flax_slots_dataclass
class Fp8Numerics(numerics.AqtNumerics):
  """Numerics for fp8."""

  dtype: Literal[jnp.float8_e4m3fn, jnp.float8_e5m2]
  exponent_bits: int = 4
  mantissa_bits: int = 3
  noise_fn: Optional[stochastic_rounding.NoiseFn] = None

  def _get_edge_of_last_fp8_bucket(self):
    return jnp.finfo(self.dtype).max.astype(jnp.bfloat16)

  def get_dtype(self):
    return self.dtype

  def abs_val_mapped_to(self):
    return self._get_edge_of_last_fp8_bucket()

  def vjp_fwd(self, x, context):
    res = (x,)
    if not (
        (self.exponent_bits == 4 and self.mantissa_bits == 3)
        or (self.exponent_bits == 5 and self.mantissa_bits == 2)
    ):
      raise ValueError(
          '(exponent_bits, mantissa_bits) can only be (4,3) or (5,2) but was '
          f'({self.exponent_bits}, {self.mantissa_bits})'
      )

    if self.noise_fn is not None:
      x = (x + self.noise_fn(x.shape, context.key)).astype(x.dtype)

    # clip
    fwd_clip_bound = self._get_edge_of_last_fp8_bucket()
    x = jnp.clip(x, -1 * fwd_clip_bound, fwd_clip_bound)

    # round
    x = round_to_nearest_even(x, self.dtype)

    return x, res

  def vjp_bwd(self, res, grad):
    # This is gradient of clip.
    # For boundary values we will have full gradient.
    # We might use something like this for calibrations other than abs(max(x))
    # (x,) = res
    # ret = (x <= edge_of_last_bucket) * (x >= -edge_of_last_bucket) * grad
    del res
    ret = grad
    return (ret, None)


def round_to_nearest_even(x: jnp.ndarray, dtype: jnp.dtype) -> jnp.ndarray:
  x = x.astype(dtype)
  # bitcast_convert to uint8 to avoid allow_excess_precision set in XLA
  x = jax.lax.bitcast_convert_type(x, jnp.uint8)
  x = jax.lax.bitcast_convert_type(x, dtype)
  return x



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration dataclasses."""


import copy
from typing import Literal, Optional, TypeAlias, Union
from aqt.jax.v2 import aqt_quantizer
from aqt.jax.v2 import calibration
from aqt.jax.v2 import stochastic_rounding

# Temporary re-export from aqt.jax.v2.aqt_dot_general
# pylint: disable=g-importing-member
# pylint: disable=unused-import
from aqt.jax.v2.aqt_dot_general import assert_config_validity
from aqt.jax.v2.aqt_dot_general import CalibrationMode
from aqt.jax.v2.aqt_dot_general import conv_general_dilated_make
from aqt.jax.v2.aqt_dot_general import DequantMode
from aqt.jax.v2.aqt_dot_general import dot_general_make
from aqt.jax.v2.aqt_dot_general import dot_general_raw_make
from aqt.jax.v2.aqt_dot_general import DotGeneral
from aqt.jax.v2.aqt_dot_general import DotGeneralRaw
from aqt.jax.v2.aqt_dot_general import dtypes_allowed_for_int32_accum
from aqt.jax.v2.aqt_dot_general import LocalAqt
from aqt.jax.v2.aqt_dot_general import Tensor
from aqt.jax.v2.aqt_dot_general import tensor_make

from aqt.jax.v2.numerics import fp8_numerics
from aqt.jax.v2.numerics import int_numerics
from aqt.jax.v2.numerics import no_numerics
from aqt.jax.v2.numerics import numerics
import jax
import jax.numpy as jnp

################################################################################
# Functions below are auxiliary config attribute setters.

# SKIP can be used as an argument to some set_xyz functions.
# It signals that set_xyz should make no changes to that option.
SKIP = 'skip'
SkipT: TypeAlias = Literal[SKIP]


def infer_dtype_from_bits(bits: int) -> jnp.dtype | None:
  """Get the dtype for the number of bits provided.

  Args:
    bits: number of bits for the dtype.

  Returns:
    The corresponding container dtype for the number of bits provided.
  """
  if bits == 4:
    # this branch should return jnp.int4 directly but
    # lax.dot_general(int4, int4) is illegal on cpu.
    # TODO(aqt): Remove this platform check once
    # https://github.com/google/jax/issues/19682 is fixed.
    if jax.local_devices()[0].platform != 'cpu':
      return jnp.int4
    else:
      return jnp.int8
  else:
    if bits <= 8 and bits >= 2:
      return jnp.int8
    else:
      return None


def _split_key(key: Optional[jax.Array], num_splits: int):
  default = (None,) * num_splits
  return default if key is None else jax.random.split(key, num_splits)


def set_context(
    cfg: DotGeneral, key: Optional[jax.Array], train_step: Optional[int]
):
  """Set context with prng keys and train_steps for dot_general config."""

  def set_dg_raw_context(cfg_raw: DotGeneralRaw, key: Optional[jax.Array]):
    key1, key2 = _split_key(key, num_splits=2)
    cfg_raw.lhs.quantizer.context = aqt_quantizer.Context(
        key=key1, train_step=train_step
    )
    cfg_raw.rhs.quantizer.context = aqt_quantizer.Context(
        key=key2, train_step=train_step
    )

  key_fwd, key_dlhs, key_drhs = _split_key(key, num_splits=3)
  ret_cfg = copy.deepcopy(cfg)
  set_dg_raw_context(ret_cfg.fwd, key_fwd)
  set_dg_raw_context(ret_cfg.dlhs, key_dlhs)
  set_dg_raw_context(ret_cfg.drhs, key_drhs)
  return ret_cfg


def set_fwd_dequant_mode(
    cfg: DotGeneral,
    *,
    lhs_dequant_mode: Optional[DequantMode] = None,
    rhs_dequant_mode: Optional[DequantMode] = None,
):
  if lhs_dequant_mode is not None:
    cfg.fwd.lhs.dequant_mode = lhs_dequant_mode
  if rhs_dequant_mode is not None:
    cfg.fwd.rhs.dequant_mode = rhs_dequant_mode


def set_fwd_calibration_mode(
    cfg: DotGeneral,
    *,
    lhs_calibration_mode: CalibrationMode | SkipT = SKIP,
    rhs_calibration_mode: CalibrationMode | SkipT = SKIP,
):
  if lhs_calibration_mode != SKIP:
    cfg.fwd.lhs.calibration_mode = lhs_calibration_mode
  if rhs_calibration_mode != SKIP:
    cfg.fwd.rhs.calibration_mode = rhs_calibration_mode


def set_numerics(
    cfg: DotGeneralRaw,
    lhs_numerics: numerics.AqtNumerics,
    rhs_numerics: numerics.AqtNumerics,
):
  """Set numerics for DotGeneralRaw config."""
  cfg.lhs.quantizer.numerics = lhs_numerics
  cfg.rhs.quantizer.numerics = rhs_numerics
  if (
      lhs_numerics.get_dtype() in dtypes_allowed_for_int32_accum
      and rhs_numerics.get_dtype() in dtypes_allowed_for_int32_accum
  ):
    cfg.dg_accumulator_dtype = jnp.int32
  elif (
      lhs_numerics.get_dtype() in fp8_numerics.fp8_map.values()
      or rhs_numerics.get_dtype() in fp8_numerics.fp8_map.values()
  ):
    cfg.dg_accumulator_dtype = jnp.float32
  else:
    cfg.dg_accumulator_dtype = None


def set_accumulator_dtype(
    cfg: DotGeneral,
    fwd_dtype: Union[jnp.dtype, None, SkipT],
    dlhs_dtype: Union[jnp.dtype, None, SkipT],
    drhs_dtype: Union[jnp.dtype, None, SkipT],
):
  if fwd_dtype != SKIP:
    cfg.fwd.dg_accumulator_dtype = fwd_dtype
  if dlhs_dtype != SKIP:
    cfg.dlhs.dg_accumulator_dtype = dlhs_dtype
  if drhs_dtype != SKIP:
    cfg.drhs.dg_accumulator_dtype = drhs_dtype


def set_stochastic_rounding(
    cfg: DotGeneral,
    # Typically we have (but it's a caller's responsibility to check):
    # - vjp_lhs_stochastic_rounding is referring to the gradient and
    # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
    vjp_lhs_stochastic_rounding: bool,
    vjp_rhs_stochastic_rounding: bool,
    implementation: str,
):
  """Configure stochastic rounding implementation."""
  noise_implementations = {
      'jax.uniform': stochastic_rounding.JaxUniform(),
      'custom-1': stochastic_rounding.RandomCenteredUniform(),
  }
  msg = f'{implementation} not supported.'
  assert implementation in noise_implementations.keys(), msg
  noise_fn = noise_implementations[implementation]

  if vjp_lhs_stochastic_rounding:
    cfg.dlhs.lhs.quantizer.numerics = cfg.dlhs.lhs.quantizer.numerics.replace(
        noise_fn=noise_fn
    )
    cfg.drhs.lhs.quantizer.numerics = cfg.drhs.lhs.quantizer.numerics.replace(
        noise_fn=noise_fn
    )
  else:
    cfg.dlhs.lhs.quantizer.numerics = cfg.dlhs.lhs.quantizer.numerics.replace(
        noise_fn=None
    )
    cfg.drhs.lhs.quantizer.numerics = cfg.drhs.lhs.quantizer.numerics.replace(
        noise_fn=None
    )

  if vjp_rhs_stochastic_rounding:
    cfg.dlhs.rhs.quantizer.numerics = cfg.dlhs.rhs.quantizer.numerics.replace(
        noise_fn=noise_fn
    )
    cfg.drhs.rhs.quantizer.numerics = cfg.drhs.rhs.quantizer.numerics.replace(
        noise_fn=noise_fn
    )
  else:
    cfg.dlhs.rhs.quantizer.numerics = cfg.dlhs.rhs.quantizer.numerics.replace(
        noise_fn=None
    )
    cfg.drhs.rhs.quantizer.numerics = cfg.drhs.rhs.quantizer.numerics.replace(
        noise_fn=None
    )


def set_static_bound(cfg: DotGeneral, bound: float = 1.0):
  cfg.fwd.lhs.quantizer.calibration = calibration.ConstantCalibration(bound)
  cfg.fwd.rhs.quantizer.calibration = calibration.ConstantCalibration(bound)
  cfg.drhs.lhs.quantizer.calibration = calibration.ConstantCalibration(bound)
  cfg.drhs.rhs.quantizer.calibration = calibration.ConstantCalibration(bound)
  cfg.dlhs.lhs.quantizer.calibration = calibration.ConstantCalibration(bound)
  cfg.dlhs.rhs.quantizer.calibration = calibration.ConstantCalibration(bound)


def set_local_aqt(
    cfg: DotGeneral,
    fwd_local_aqt: Union[SkipT, LocalAqt, None],
    dlhs_local_aqt: Union[SkipT, LocalAqt, None],
    drhs_local_aqt: Union[SkipT, LocalAqt, None],
):
  if fwd_local_aqt != SKIP:
    cfg.fwd.local_aqt = fwd_local_aqt
  if dlhs_local_aqt != SKIP:
    cfg.dlhs.local_aqt = dlhs_local_aqt
  if drhs_local_aqt != SKIP:
    cfg.drhs.local_aqt = drhs_local_aqt


def set_use_fwd_quant(
    cfg: DotGeneral,
    dlhs_use_fwd_quant: Union[bool, None, SkipT],
    drhs_use_fwd_quant: Union[bool, None, SkipT],
):
  if dlhs_use_fwd_quant != SKIP:
    cfg.dlhs.rhs.use_fwd_quant = dlhs_use_fwd_quant
  if drhs_use_fwd_quant != SKIP:
    cfg.drhs.rhs.use_fwd_quant = drhs_use_fwd_quant


def set_int_numerics_preserve_zero(cfg: DotGeneral, preserve_zero: bool):
  for dot_general_raw in [cfg.fwd, cfg.dlhs, cfg.drhs]:
    for tensor in [dot_general_raw.lhs, dot_general_raw.rhs]:
      if isinstance(tensor.quantizer.numerics, int_numerics.IntNumerics):
        tensor.quantizer.numerics.preserve_zero = preserve_zero
        updated_dtype = (
            infer_dtype_from_bits(tensor.quantizer.numerics.bits)
            if preserve_zero
            else None
        )
        tensor.quantizer.numerics.dtype = updated_dtype


def set_absmax_calib_scale(cfg: DotGeneral, scale: float):
  """Set AbsMaxCalibration scale and update clip_gradient accordingly."""
  for dot_general_raw in [cfg.fwd, cfg.dlhs, cfg.drhs]:
    for tensor in [dot_general_raw.lhs, dot_general_raw.rhs]:
      assert isinstance(
          tensor.quantizer.calibration, calibration.AbsMaxCalibration
      ), (
          'scale is only available in AbsMaxCalibration, while'
          f'{tensor.quantizer.calibration} is used in current config.'
      )
      tensor.quantizer.calibration = calibration.AbsMaxCalibration(scale)
      if scale < 1.0 and isinstance(
          tensor.quantizer.numerics, int_numerics.IntNumerics
      ):
        tensor.quantizer.numerics.clip_gradient = True


def set_bits(
    cfg: DotGeneral,
    fwd_lhs_bit: Union[int, None, fp8_numerics.FP8Dtype],
    fwd_rhs_bit: Union[int, None, fp8_numerics.FP8Dtype],
    dlhs_lhs_bit: Union[int, None, fp8_numerics.FP8Dtype],
    dlhs_rhs_bit: Union[int, None, fp8_numerics.FP8Dtype],
    drhs_lhs_bit: Union[int, None, fp8_numerics.FP8Dtype],
    drhs_rhs_bit: Union[int, None, fp8_numerics.FP8Dtype],
) -> DotGeneral:
  """Set quantization bits for dot_general config."""

  def get_numerics(bits):
    if bits is None:
      effective_numerics = no_numerics.NoNumerics()
    elif bits in fp8_numerics.fp8_map.keys():
      exponent_bits, mantissa_bits = int(bits[1]), int(bits[3])
      effective_numerics = fp8_numerics.Fp8Numerics(
          exponent_bits=exponent_bits,
          mantissa_bits=mantissa_bits,
          dtype=fp8_numerics.fp8_map[bits],
          noise_fn=None,
      )
    else:
      pz = False if bits == 1 else True
      dtype = infer_dtype_from_bits(bits) if pz else None
      effective_numerics = int_numerics.IntNumerics(
          bits=bits,
          preserve_zero=pz,
          preserve_max_val=False,
          clip=True,
          round=True,
          noise_fn=None,
          clip_gradient=False,  # Can be disabled when using abs-max scaling.
          dtype=dtype,
      )
    return effective_numerics

  set_numerics(cfg.fwd, get_numerics(fwd_lhs_bit), get_numerics(fwd_rhs_bit))
  set_numerics(cfg.dlhs, get_numerics(dlhs_lhs_bit), get_numerics(dlhs_rhs_bit))
  set_numerics(cfg.drhs, get_numerics(drhs_lhs_bit), get_numerics(drhs_rhs_bit))
  # use_fwd_quant is by default set to False if fwd pass is quantized.
  # This is to make the configuration logically correct,
  # i.e., use_fwd_quant cannot be None when fwd is quantized.
  # It is user's responsibility to further choose between False and True.
  dlhs_use_fwd_quant = False if fwd_rhs_bit is not None else SKIP
  drhs_use_fwd_quant = False if fwd_lhs_bit is not None else SKIP
  set_use_fwd_quant(cfg, dlhs_use_fwd_quant, drhs_use_fwd_quant)
  return cfg


################################################################################
# Functions below are auxiliary config creators.


def default_unquantized_config() -> DotGeneral:
  """Aqt config for floating-point dot general."""

  def tensor_cfg() -> Tensor:
    quantizer = aqt_quantizer.Quantizer(
        numerics=no_numerics.NoNumerics(),
        calib_shared_axes=None,
        scale_stop_grad=True,
        calibration=calibration.AbsMaxCalibration(),
        po2_scale=False,
        context=aqt_quantizer.Context(key=None, train_step=None),
    )
    cfg = Tensor(
        quantizer=quantizer,
        use_fwd_quant=None,
        dequant_mode=DequantMode.OUTPUT,
        calibration_mode=CalibrationMode.CONTRACTING_AXIS,
    )
    return cfg

  def dg_raw_cfg(jax_scope_name: str) -> DotGeneralRaw:
    return DotGeneralRaw(
        lhs=tensor_cfg(),
        rhs=tensor_cfg(),
        dg_accumulator_dtype=None,
        local_aqt=None,
        jax_scope_name=jax_scope_name,
    )

  dg_cfg = DotGeneral(
      fwd=dg_raw_cfg('aqt_fwd'),
      dlhs=dg_raw_cfg('aqt_dlhs'),
      drhs=dg_raw_cfg('aqt_drhs'),
  )
  return dg_cfg


def fully_quantized(
    *,
    fwd_bits: Optional[int] = 8,
    bwd_bits: Optional[int] = 8,
    use_fwd_quant: bool = True,
    use_stochastic_rounding: Optional[bool] = True,
    # Typically we have (but it's a caller's responsibility to check):
    # - vjp_lhs_stochastic_rounding is referring to the gradient and
    # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
    vjp_lhs_stochastic_rounding: Optional[bool] = None,
    vjp_rhs_stochastic_rounding: Optional[bool] = None,
    # The dummy static bound flag is temporary, for performance benchmarking.
    use_dummy_static_bound: bool = False,
    dlhs_local_aqt: Optional[LocalAqt] = None,
    drhs_local_aqt: Optional[LocalAqt] = None,
) -> DotGeneral:
  """Fully Quantized Training."""
  cfg = dot_general_make(
      lhs_bits=fwd_bits,
      rhs_bits=fwd_bits,
      bwd_bits=bwd_bits,
      use_fwd_quant=use_fwd_quant,
      dlhs_local_aqt=dlhs_local_aqt,
      drhs_local_aqt=drhs_local_aqt,
  )

  # Stochastic Rounding
  # These 3 variables are used to ensure we don't mix
  # old and new style of SR configuration.
  old_style_sr_config = use_stochastic_rounding is not None
  new_style_sr_config_lhs = vjp_lhs_stochastic_rounding is not None
  new_style_sr_config_rhs = vjp_rhs_stochastic_rounding is not None
  assert new_style_sr_config_lhs == new_style_sr_config_rhs, (
      'if you use new style SR config (vjp_xhs_stochastic_rounding), do pass'
      ' both lhs and rhs explicitely.'
  )
  assert new_style_sr_config_lhs != old_style_sr_config

  true = True  # A crude way to get around g-explicit-bool-comparison warning

  assert not (vjp_lhs_stochastic_rounding and vjp_rhs_stochastic_rounding), (
      'This config is buggy when you set both to True. Contact lew@ or use'
      ' config_v3'
  )

  # By default use jax.uniform for stochastic rounding
  if use_stochastic_rounding == true:
    set_stochastic_rounding(cfg, True, True, 'jax.uniform')

  if vjp_lhs_stochastic_rounding == true:
    set_stochastic_rounding(cfg, True, False, 'jax.uniform')

  if vjp_rhs_stochastic_rounding == true:
    set_stochastic_rounding(cfg, False, True, 'jax.uniform')

  if use_dummy_static_bound:
    set_static_bound(cfg, 1.0)

  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'

  return cfg


def config_v3(
    *,
    fwd_bits: Optional[int] = 8,
    dlhs_bits: Optional[int] = 8,
    drhs_bits: Optional[int] = None,
    # The dummy static bound flag is for performance benchmarking.
    use_dummy_static_bound: bool = False,
    rng_type: str = 'jax.uniform',  # 'custom-1'
    dlhs_local_aqt: Optional[LocalAqt] = None,
    drhs_local_aqt: Optional[LocalAqt] = None,
    fwd_accumulator_dtype: ... = jnp.int32,
    dlhs_accumulator_dtype: ... = jnp.int32,
    drhs_accumulator_dtype: ... = None,
) -> DotGeneral:
  """Fully Quantized Training."""
  fwd = dot_general_raw_make(fwd_bits, fwd_bits, jax_scope_name='aqt_fwd')
  dlhs = dot_general_raw_make(
      dlhs_bits,
      dlhs_bits,
      local_aqt=dlhs_local_aqt,
      jax_scope_name='aqt_dlhs',
  )
  drhs = dot_general_raw_make(
      drhs_bits,
      drhs_bits,
      local_aqt=drhs_local_aqt,
      jax_scope_name='aqt_drhs',
  )
  cfg = DotGeneral(fwd=fwd, dlhs=dlhs, drhs=drhs)

  cfg.dlhs.rhs.use_fwd_quant = False
  cfg.drhs.rhs.use_fwd_quant = False

  # Typically we have (but I don't know if it is guraranteed):
  # - vjp_lhs_stochastic_rounding is referring to the gradient and
  # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
  set_stochastic_rounding(
      cfg,
      vjp_lhs_stochastic_rounding=True,
      vjp_rhs_stochastic_rounding=False,
      implementation=rng_type,
  )

  if use_dummy_static_bound:
    set_static_bound(cfg, 1.0)

  set_accumulator_dtype(
      cfg,
      fwd_dtype=fwd_accumulator_dtype,
      dlhs_dtype=dlhs_accumulator_dtype,
      drhs_dtype=drhs_accumulator_dtype,
  )
  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'
  return cfg


def config_v4(
    *,
    fwd_bits: Optional[int] = 8,
    dlhs_bits: Optional[int] = 8,
    drhs_bits: Optional[int] = None,
    # The dummy static bound flag is for performance benchmarking.
    use_dummy_static_bound: bool = False,
    rng_type: str = 'jax.uniform',  # 'custom-1'
    dlhs_local_aqt: Optional[LocalAqt] = None,
    drhs_local_aqt: Optional[LocalAqt] = None,
    # accumulator dtype by default is automatically set in set_bits,
    # but users can still configure a special dtype such as jnp.int16, etc.
    fwd_accumulator_dtype: Union[jnp.dtype, None, SkipT] = SKIP,
    dlhs_accumulator_dtype: Union[jnp.dtype, None, SkipT] = SKIP,
    drhs_accumulator_dtype: Union[jnp.dtype, None, SkipT] = SKIP,
) -> DotGeneral:
  """Version 4 of user-visible AQT config."""
  cfg = default_unquantized_config()
  set_bits(
      cfg,
      fwd_lhs_bit=fwd_bits,
      fwd_rhs_bit=fwd_bits,
      dlhs_lhs_bit=dlhs_bits,
      dlhs_rhs_bit=dlhs_bits,
      drhs_lhs_bit=drhs_bits,
      drhs_rhs_bit=drhs_bits,
  )
  set_accumulator_dtype(
      cfg,
      fwd_dtype=fwd_accumulator_dtype,
      dlhs_dtype=dlhs_accumulator_dtype,
      drhs_dtype=drhs_accumulator_dtype,
  )
  set_stochastic_rounding(
      cfg,
      vjp_lhs_stochastic_rounding=True,
      vjp_rhs_stochastic_rounding=False,
      implementation=rng_type,
  )
  if use_dummy_static_bound:
    set_static_bound(cfg, 1.0)
  set_local_aqt(
      cfg,
      fwd_local_aqt=SKIP,
      dlhs_local_aqt=dlhs_local_aqt,
      drhs_local_aqt=drhs_local_aqt,
  )
  # TODO(yichizh): remove set_use_fwd_quant here since it will be automatically
  # set in set_bits. Or make them as an argument.
  set_use_fwd_quant(cfg, dlhs_use_fwd_quant=False, drhs_use_fwd_quant=False)
  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'
  return cfg


def config_fwd_fp8(fwd_bits: fp8_numerics.FP8Dtype = 'e4m3') -> DotGeneral:
  """Configs for FP8 forward pass."""
  assert (
      fwd_bits in fp8_numerics.fp8_map.keys()
  ), 'FP8 only supports 4 or 5 exponent bits'
  cfg = config_v4(fwd_bits=8, dlhs_bits=None, drhs_bits=None)
  set_bits(
      cfg,
      fwd_lhs_bit=fwd_bits,
      fwd_rhs_bit=fwd_bits,
      dlhs_lhs_bit=None,
      dlhs_rhs_bit=None,
      drhs_lhs_bit=None,
      drhs_rhs_bit=None,
  )
  set_stochastic_rounding(cfg, False, False, 'jax.uniform')
  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'
  return cfg



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Full-sized unquantized model.

A 'full size' means
* 6 layers in the encoder and decoder
* 1024 emd_dim
* 16 heads
* 1024 qkv_dim
* 4096 mlp_dim
"""

from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs import base_config


def get_config(quant_target=base_config.QuantTarget.NONE):
  config = base_config.get_config(n_layers=6, quant_target=quant_target)
  config.metadata.hyper_str = 'full_bfloat16'
  return config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for aqt_tensor."""

from absl.testing import absltest
from aqt.jax import aqt_tensor
from aqt.test import aqt_stats_test_base
from aqt.test import aqt_tensor_test_base
import jax
import jax.numpy as jnp
import numpy as np


def f32(x):
  """Casts 'x' to be a float32 jax array."""
  return jnp.array(x, dtype=jnp.float32)


class StatsTest(aqt_stats_test_base.StatsTest):
  """Tests for Stats class.

  Refer to aqt_stats_test_base.StatsTest for more details.
  """

  def set_stats(self, data_shape, config):
    self._stats = aqt_tensor.Stats(data_shape=data_shape, config=config)
    self._stats_state = self._stats.init(jax.random.PRNGKey(0))
    self.override_ema_update_count = None

  def update(self, sample, weight):
    _, self._stats_state = self._stats.apply(
        self._stats_state,
        f32(sample),
        f32(weight),
        override_ema_update_count=self.override_ema_update_count,
        method=self._stats.update,
        mutable=["aqt"])

  def get_sum_of_ones(self):
    return self.stats_state["aqt"]["sum_of_ones"]

  def get_sum_of_vals(self):
    return self.stats_state["aqt"]["sum_of_vals"]

  def get_max_of_abs_vals(self):
    return self.stats_state["aqt"]["max_of_abs_vals"]

  def get_sum_of_l1_vals(self):
    return self.stats_state["aqt"]["sum_of_l1_vals"]

  def get_sum_of_lp_vals(self):
    return self.stats_state["aqt"]["sum_of_lp_vals"]

  def set_ema_update_count(self, ema_update_count):
    self.override_ema_update_count = ema_update_count

  def mean(self):
    return self.stats.apply(self.stats_state, method=self.stats.mean)

  def max_dev(self):
    return self.stats.apply(self.stats_state, method=self.stats.max_dev)

  def l1_dev(self):
    return self.stats.apply(self.stats_state, method=self.stats.l1_dev)

  def lp_dev(self):
    return self.stats.apply(self.stats_state, method=self.stats.lp_dev)

  def bound(self, calibration_config):
    return self.stats.apply(
        self.stats_state, calibration_config, method=self.stats.bound
    )


class AqtTensorQuantizerTest(aqt_tensor_test_base.AqtTensorQuantizerTest):
  """Tests for AqtTensorQuantizer class.

  Refer to aqt_test_shared_base.AqtTensorQuantizerTest for more details.
  """

  _quant_state = {}

  def make_tensor_quantizer(self, data_shape, config, name="tq"):
    quant = aqt_tensor.TensorQuantizer(
        data_shape=data_shape, config=config, name=name)
    self._quant_state[name] = quant.init(jax.random.PRNGKey(0))
    return quant

  def update_quantizer(self, quant, sample, weight, event_count):
    _, self._quant_state[quant.name] = quant.apply(
        self._quant_state[quant.name],
        sample,
        weight,
        int(event_count),
        method=quant.update,
        mutable=["TensorQuantizer", "aqt"])

  def to_quant(self, quant, x, train=True):
    return quant.apply(
        self._quant_state[quant.name], x, train, method=quant._to_quant)

  def get_quant_scale(self, quant, train=True):
    return quant.apply(
        self._quant_state[quant.name],  #
        train,
        method=quant._get_quant_scale)

  def init(self):
    pass

  def get_scale(self, quant):
    return self._quant_state[quant.name]["TensorQuantizer"]["scale"]

  def get_last_update(self, quant):
    return self._quant_state[quant.name]["TensorQuantizer"]["last_update"]

  def get_clip_range(self, quant):
    return quant.apply(self._quant_state[quant.name], method=quant.clip_range)

  def get_quantized_variable(self, quant):
    return self._quant_state[
        quant.name]["TensorQuantizer"]["quantized_variable"]

  def test_pass_through(self):
    inp = jnp.float32(100)
    eps = jnp.finfo(jnp.float32).eps

    def fn(x):
      return x * eps + eps

    actual = aqt_tensor.pass_through(inp, fn)
    expected = fn(inp)

    # Pass-through function should return the same output as fn(x).
    np.testing.assert_equal(actual, expected)

    # The gradient of fn() should be 1.0 since STE makes it pretend to be an
    # identity function during the backward pass.
    np.testing.assert_equal(
        jnp.array(1.0),
        jax.grad(aqt_tensor.pass_through)(inp, fn))


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for conv2d."""

import itertools

from absl.testing import parameterized
from aqt.common import aqt_config
from aqt.test import aqt_test_shared_base
import numpy as np
import tensorflow.compat.v1 as tf


def input_stats(filter_zeros=True):
  """Generates stats configuration for default NHWC inputs."""
  return aqt_config.StatsConfig(
      ema_update_count=1,
      share_stats_axes=[1, 2, 3],
      update_count_prior=0,
      tpu_cross_replica_sum=False,
      filter_zeros=filter_zeros)


def filter_stats(filter_zeros=True):
  """Generates stats configuration for default HWIO filter."""
  sc = input_stats(filter_zeros)
  sc.share_stats_axes = [0, 1, 2]
  return sc


def int_quant_config(bits, preserve_zero=True):
  """Returns a zero-preserving int config with the given bits."""
  return aqt_config.IntQuantConfig(bits=bits, preserve_zero=preserve_zero)


def calibration_config(const_coeff, max_dev_coeff):
  """Returns a calibration config with the given constant coeff."""
  return aqt_config.CalibrationConfig(
      const_bound_coeff=const_coeff, max_dev_coeff=max_dev_coeff)


def stats_config(input_or_filter, filter_zeros=True):
  assert input_or_filter in ["input", "filter"]
  stats_config_fn = input_stats if input_or_filter == "input" else filter_stats
  return stats_config_fn(filter_zeros)


def schedule_config(input_or_filter,
                    *,
                    const_coeff,
                    bits,
                    max_dev_coeff=0,
                    filter_zeros=True,
                    preserve_zero=True):
  """Combines all helpers above to set schedule config."""
  tensor_config = aqt_config.AqtTensorConfig(
      quant_config=int_quant_config(bits, preserve_zero),
      calibration_config=calibration_config(const_coeff, max_dev_coeff),
      freeze_scale_at_begin=True)
  sc = stats_config(input_or_filter, filter_zeros)
  return aqt_config.AqtScheduleConfig(sc, [tensor_config])


def empty_config(input_or_filter):
  return aqt_config.AqtScheduleConfig(stats_config(input_or_filter), [])


def generate_missing_contraction_dims():
  """Cases where some spatial dims for conv are missing."""

  cases = []

  data_format = ["NHWC"]
  correct_filter_cases = [("correct", [0, 1, 2])]
  correct_input_cases = [("correct", [1, 2, 3])]
  incorrect_filter_cases = [("partial", [1]), ("wrong", [1, 3]),
                            ("partial2", [1, 2]), ("partial3", [0, 2]),
                            ("missing", [])]
  incorrect_input_cases = [("partial", [2]), ("wrong", [0, 1]),
                           ("partial2", [1, 2]), ("missing", [])]

  def add_cases(input_cases, filter_cases):
    case_tuples = itertools.product(data_format, input_cases, filter_cases)
    cases.extend(case_tuples)

  add_cases(correct_input_cases, incorrect_filter_cases)
  add_cases(incorrect_input_cases, correct_filter_cases)
  add_cases(incorrect_input_cases, incorrect_filter_cases)

  data_format = ["NCHW"]
  correct_input_cases = [("correct", [1, 2, 3])]
  incorrect_input_cases = [("partial", [2]), ("wrong", [0, 1]),
                           ("partial2", [1, 2]), ("missing", [])]

  add_cases(correct_input_cases, incorrect_filter_cases)
  add_cases(incorrect_input_cases, correct_filter_cases)
  add_cases(incorrect_input_cases, incorrect_filter_cases)

  case_dicts = []
  for data_format, (i_name, input_axes), (f_name, filter_axes) in cases:
    case_dicts.append({
        "data_format": data_format,
        "input_axes": input_axes,
        "filter_axes": filter_axes,
        "testcase_name": f"{data_format}_input_{i_name}_filter_{f_name}"
    })

  return case_dicts


class ConvTest(tf.test.TestCase, parameterized.TestCase):
  """Base class for testing aqt_conv2d in TF and aqt_conv_general in Jax."""

  def setUp(self):
    """Seed random for deterministic but nontrivial inputs."""
    super().setUp()
    self.rng = np.random.default_rng(1234)

  def create_random_input_and_filter(self,
                                     input_shape=(1, 3, 3, 1),
                                     filter_shape=(2, 2, 1, 2)):
    x = self.rng.standard_normal(size=input_shape, dtype=np.float32)
    w = self.rng.standard_normal(size=filter_shape, dtype=np.float32)
    return x, w

  def exact_int8_conv_example(self,
                              lhs_shape=(1, 3, 3, 1),
                              rhs_shape=(2, 2, 1, 2),
                              lhs_share_stats_axes=(1, 2, 3),
                              rhs_share_stats_axes=(0, 1, 2),
                              lhs_use_quantized_variable=False,
                              rhs_use_quantized_variable=False):
    return aqt_test_shared_base.exact_int8_example(
        lhs_shape=lhs_shape,
        rhs_shape=rhs_shape,
        lhs_share_stats_axes=list(lhs_share_stats_axes),
        rhs_share_stats_axes=list(rhs_share_stats_axes),
        lhs_use_quantized_variable=lhs_use_quantized_variable,
        rhs_use_quantized_variable=rhs_use_quantized_variable)

  def conv_op_quantized(
      self,
      input,  # pylint: disable=redefined-builtin
      filter,  # pylint: disable=redefined-builtin
      input_config,
      filter_config,
      event_count,
      input_weights=None,
      train=True,
      **kwargs):
    """Performs quantized convolution op.

    For given input and filter configs, this method creates TensorQuantizers,
    update them to the event_count, and returns the result of quantized
    convolution. The reason to deal with TensorQuantizer inside thie method is
    that TensorQuantizer in Jax is nn.Module that should be initialized and
    updated inside another nn.Module. Thus, the arguments needed to manage
    TensorQuantizers are passed and then how they are handled needs to be
    defined in the inheriting classes.

    Args:
      input: The convolution input (lhs).
      filter: The convolution kernel (rhs).
      input_config: AqtScheduleConfig for the input.
      filter_config: AqtScheduleConfig for the filter.
      event_count: event_count to which input/filter TensorQuantizers should be
        updated
      input_weights: The weight of input for updating statistics.
      train: Indicates training mode if True, otherwise, inference mode.
      **kwargs: Keyword arguments to pass onto conv op.
    Returns:
      A tensor of quantized convolution result.
    """
    raise NotImplementedError

  def conv_op_unquantized(self, input, filter, **kwargs):  # pylint: disable=redefined-builtin
    """Performs unquantized convolution op such as tf.conv2d and lax.conv_general_dilated."""
    raise NotImplementedError

  def get_conv_kwargs(
      self,
      strides,
      padding,
      data_format="NHWC",
      dilations=None):
    """Returns a dict of keyword arguments for convolution.

    tf.conv2d and lax.conv_general_dilated have different argument sets,
    names, and their types. For example, the stride of a convolution is
    determined by `strides` in tf.conv2d but by `window_strides` in
    lax.conv_general_dilated. This wrapper should return a dict of such keyword
    arguments customized for each framework.

    Args:
      strides: Stride of the convolution.
      padding: Padding added to all four sides of the input.
      data_format: Data format of the input/output.
      dilations: Spacing between kernel elements.
    """
    raise NotImplementedError

  def get_module_and_side_effect(self):
    raise NotImplementedError

  def constant(self, x):
    raise NotImplementedError

  def gradients(self, fwd_func, x, w):
    raise NotImplementedError

  def test_basic_conv(self):
    """Test on a hand-computed example."""
    x = np.array([[1.3, -0.3, -1.1], [-2.3, 1.1, 0.0], [0.0, -0.5, 10]],
                 dtype=np.float32)[np.newaxis, :, :, np.newaxis]
    w0 = np.array([[1, 5.1], [-5.5, -3]], dtype=np.float32)
    w1 = np.array([[50, 6], [-1.1, 1.2]], dtype=np.float32)
    w = np.stack([w0, w1])[:, :, np.newaxis, :]

    # choose a const such that the clip is 2 for 3 buckets
    cutoff = 2 * 1.5
    input_config = schedule_config("input", const_coeff=cutoff, bits=2)

    cutoff = 10 * 1.5
    filter_config = schedule_config("filter", const_coeff=cutoff, bits=2)

    kwargs = self.get_conv_kwargs(strides=1, padding="VALID")
    actual = self.conv_op_quantized(
        x,
        w,
        input_config,
        filter_config,
        event_count=0,
        **kwargs,
    )

    x = np.array([[1, 0, -1], [-1, 1, 0], [0, 0, 1]],
                 dtype=np.float32)[np.newaxis, :, :, np.newaxis]
    w0 = np.array([[0, 1], [-1, 0]], dtype=np.float32)
    w1 = np.array([[1, 1], [0, 0]], dtype=np.float32)
    w = np.stack([w0, w1])[:, :, np.newaxis, :]

    expected = self.conv_op_unquantized(x, w, **kwargs)
    expected *= 2
    expected *= 10

    self.assertAllEqual(expected, actual)

  @parameterized.named_parameters(generate_missing_contraction_dims())
  def test_missing_contraction_dims(self, data_format, filter_axes, input_axes):
    """Throw if spatial dimension stats are unshared."""

    input_config = schedule_config("input", bits=8, const_coeff=1.0)
    input_config.stats_config.share_stats_axes[:] = input_axes
    filter_config = schedule_config("filter", bits=8, const_coeff=1.0)
    filter_config.stats_config.share_stats_axes[:] = filter_axes

    x = np.zeros((1, 3, 3, 1), dtype=np.float32)
    w = np.zeros((2, 2, 1, 2), dtype=np.float32)

    kwargs = self.get_conv_kwargs(
        strides=1, padding="VALID", data_format=data_format)
    with self.assertRaises(aqt_config.ConfigError):
      self.conv_op_quantized(
          x,
          w,
          input_config,
          filter_config,
          event_count=0,
          **kwargs,
      )

  @parameterized.parameters(
      dict(strides=1, padding="SAME", dilations=[1, 2]),
      dict(strides=2, padding="SAME", dilations=[1, 2]),
      dict(strides=1, padding="VALID", dilations=[1, 2]),
      dict(strides=2, padding="VALID", dilations=[1, 2]),
      dict(strides=1, padding="SAME", dilations=[2, 1]),
      dict(strides=2, padding="SAME", dilations=[2, 1]),
      dict(strides=1, padding="VALID", dilations=[2, 1]),
      dict(strides=2, padding="VALID", dilations=[2, 1]),)
  def test_exact_int8_and_no_quant(self, strides, padding, dilations):
    """Validate conv results for exact_int8 and no quantization."""
    kwargs = self.get_conv_kwargs(
        strides=strides,
        padding=padding,
        data_format="NHWC",
        dilations=dilations)

    input_config, x, filter_config, w = self.exact_int8_conv_example()

    actual_exact_int8 = self.conv_op_quantized(
        x,
        w,
        input_config,
        filter_config,
        event_count=0,
        var_scope_name="exact_int8_conv",
        **kwargs,
    )
    expected_exact_int8 = self.conv_op_unquantized(x, w, **kwargs)

    self.assertAllEqual(actual_exact_int8, expected_exact_int8)

    x, w = self.create_random_input_and_filter()
    actual_no_quant = self.conv_op_quantized(
        x,
        w,
        empty_config("input"),
        empty_config("filter"),
        event_count=0,
        var_scope_name="no_quantization",
        **kwargs,
    )
    expected_no_quant = self.conv_op_unquantized(x, w, **kwargs)

    self.assertAllEqual(actual_no_quant, expected_no_quant)

  def test_exact_grads(self):
    """Ensures both quantized and unquantized conv op emit the same gradients."""
    input_config, x, filter_config, w = self.exact_int8_conv_example()

    kwargs = self.get_conv_kwargs(strides=1, padding="VALID")

    def actual_fwd(x, w):
      return self.conv_op_quantized(
          x, w, input_config, filter_config, event_count=0, **kwargs)

    def expected_fwd(x, w):
      return self.conv_op_unquantized(x, w, **kwargs)

    actual = self.gradients(actual_fwd, x, w)
    expected = self.gradients(expected_fwd, x, w)

    for actual_grad, expected_grad in zip(actual, expected):
      self.assertAllEqual(actual_grad, expected_grad)

  def test_filter_dilation(self):
    """Validates zero preservation on the filter when using filater dilation."""
    x = np.array([[1.3, -0.3, -1.1, 0.9], [-2.3, 1.1, 0.0, 1.3],
                  [0.0, -0.5, 10, -0.2], [1.2, 0.0, 5.4, -0.4]],
                 dtype=np.float32)[np.newaxis, :, :, np.newaxis]
    w0 = np.array([[1, 5.1], [-5.5, -3]], dtype=np.float32)
    w1 = np.array([[50, 6], [-1.1, 1.2]], dtype=np.float32)
    w = np.stack([w0, w1])[:, :, np.newaxis, :]
    kwargs = self.get_conv_kwargs(strides=1, padding="VALID", dilations=[2, 2])

    # choose a const such that the clip is 2 for 3 buckets
    cutoff = 2 * 1.5
    input_config = schedule_config("input", const_coeff=cutoff, bits=2)

    cutoff = 10 * 1.5
    filter_config = schedule_config("filter", const_coeff=cutoff, bits=2)

    actual_dilated = self.conv_op_quantized(
        x,
        w,
        input_config,
        filter_config,
        event_count=0,
        var_scope_name="dilated",
        **kwargs,
    )

    # Create the same filter as above but with zero values on the second row
    # and column, resulting in shape (3, 3, 1, 2). This is equivalent with the
    # (2, 2, 1, 2) shaped filter with a dilation factor of 2.
    w0 = np.array([[1, 5.1], [0.0, 0.0], [-5.5, -3]], dtype=np.float32)
    w1 = np.array([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]], dtype=np.float32)
    w2 = np.array([[50, 6], [0.0, 0.0], [-1.1, 1.2]], dtype=np.float32)
    w = np.stack([w0, w1, w2])[:, :, np.newaxis, :]

    # No dilation.
    kwargs = self.get_conv_kwargs(strides=1, padding="VALID", dilations=None)
    actual_not_dilated = self.conv_op_quantized(
        x,
        w,
        input_config,
        filter_config,
        event_count=0,
        var_scope_name="not_dilated",
        **kwargs,
    )

    x = np.array([[1, 0, -1, 0], [-1, 1, 0, 1], [0, 0, 1, 0], [1, 0, 1, 0]],
                 dtype=np.float32)[np.newaxis, :, :, np.newaxis]
    w0 = np.array([[0, 1], [-1, 0]], dtype=np.float32)
    w1 = np.array([[1, 1], [0, 0]], dtype=np.float32)
    w = np.stack([w0, w1])[:, :, np.newaxis, :]
    kwargs = self.get_conv_kwargs(strides=1, padding="VALID", dilations=[2, 2])

    expected = self.conv_op_unquantized(x, w, **kwargs)
    expected *= 2
    expected *= 10

    self.assertAllEqual(actual_dilated, actual_not_dilated)
    self.assertAllEqual(actual_dilated, expected)

  def test_zero_preservation_for_filter_dilation(self):
    """Validates preserve_zero is set to True when the filter is dilated."""
    input_config = schedule_config("input", bits=8, const_coeff=1.0)
    filter_config = schedule_config(
        "filter", bits=8, const_coeff=1.0, preserve_zero=False)

    x, w = self.create_random_input_and_filter()
    kwargs = self.get_conv_kwargs(strides=1, padding="VALID", dilations=[2, 2])

    with self.assertRaisesRegex(
        (aqt_config.ConfigError, tf.errors.InvalidArgumentError),
        "must be True if the filter is dilated"):
      self.conv_op_quantized(
          x,
          w,
          input_config,
          filter_config,
          event_count=0,
          **kwargs)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Types for configuring quantization."""

import dataclasses
import enum
import typing
from aqt.jax_legacy.jax.flax import struct as flax_struct
import flax

dataclass = flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass


class QuantGranularity(str, enum.Enum):
  """The granularity of the scale factors used for quantization."""
  # A single scale factor is used for an entire tensor.
  PER_TENSOR = 'per_tensor'
  # For a weight tensor, a separate scale factor is used for each dimension
  # corresponding to an output channel (ie, the columns of the weight matrix if
  # a dense layer is thought of as an "activation tensor * weight tensor" matrix
  # multiplication).
  PER_CHANNEL = 'per_channel'


@dataclass
class DynamicContext:
  """Dynamic parameters visible to all model layers.

  A single instance of this dataclass is intended to be threaded throughout the
  entire hierarchy of model layers, making it easy when the model is called
  (whether for training or inference) for the caller to set contextual
  variables that all parts of the model that deal with quantization need.
  """

  # The construct `field_name = flax.struct.field(pytree_node=False)` causes a
  # field to be treated as a *static* argument to a JITed function. ie, the
  # function will recompile when the value of that field changes, but normal
  # Python control flow can be used inside the model (eg, `if
  # context.update_bounds:`) with those fields. Other fields will be treated as
  # *dynamic*, so the model will not recompile when those fields change, but
  # they cannot be used in normal Python control flow (typically, you would use
  # `lax.cond` instead).

  # Whether to update activation bounds.
  update_bounds: bool = flax.struct.field(pytree_node=False)

  # This will be passed to create_weight_ops in ConvAqt
  quantize_weights: bool = flax.struct.field(default=True, pytree_node=False)

  # Whether to tag activations to record statistics.
  collect_acts_stats: bool = flax.struct.field(default=False, pytree_node=False)

  # Whether to apply sparsity for training, this is set to True somewhere in a
  # later stage of training.
  # Currently we separate the update of sparsity masks for weight and
  # activation, since during evaluation the weight sparsity mask is
  # reused from training, while activation mask will need update
  # (due to change of seq length).
  # The apply_sparsity flag is shared between weight/activation sparsity,
  # we can modify it to use two flags in the future if necessary.
  apply_sparsity: bool = flax.struct.field(default=False, pytree_node=False)
  update_weight_sparsity: bool = flax.struct.field(
      default=False, pytree_node=False)
  update_act_sparsity: bool = flax.struct.field(
      default=False, pytree_node=False)
  num_update_sparsity: int = flax.struct.field(default=0, pytree_node=False)

  # Whether to quantize activations.
  #
  # TODO(malmaud): This only applies to softmax for now. Apply it to layernorm
  # and dot operations as well.
  quantize_acts: bool = True

  # Whether to feed lax.dot inputs with an int8 dtype and accumulate to int32
  # dtype if quantizing both inputs to the dot operation to 8bits or 4bits (thus
  # ensuring the cast to int is lossless). If False, inputs are always
  # floating-point.
  prefer_int8_to_int32_dot: bool = True



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.fp_cast."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import fp_cast
import jax
from jax import config
import jax.numpy as jnp
import numpy as onp

config.update('jax_numpy_rank_promotion', 'raise')

fp8_143_max_value = float.fromhex('0x1.ep4')
fp8_143_min_value = float.fromhex('0x1p-11')


def test_data(dtype):
  return (
      dict(
          testcase_name=f'{dtype.dtype.name}_fp8_143_special',
          dtype=dtype,
          argument_result_values=[
              [float('nan'), float('nan')],
              [float('inf'), float('inf')],
              [float('-inf'), float('-inf')],
          ],
      ),
      dict(
          testcase_name=f'{dtype.dtype.name}_fp8_143_zeros',
          dtype=dtype,
          argument_result_values=[
              [-0., -0.],
              [0., 0.],
          ],
      ),
      dict(
          testcase_name=f'{dtype.dtype.name}_fp8_143_saturation',
          dtype=dtype,
          argument_result_values=[
              [fp8_143_max_value, fp8_143_max_value],
              [float.fromhex('0x1.fp4'), fp8_143_max_value],
              [float.fromhex('0x1.cp4'),
               float.fromhex('0x1.cp4')],
              [-fp8_143_max_value, -fp8_143_max_value],
              [float.fromhex('-0x1.fp4'), -fp8_143_max_value],
              [float.fromhex('-0x1.cp4'),
               float.fromhex('-0x1.cp4')],
          ],
      ),
      dict(
          testcase_name=f'{dtype.dtype.name}_fp8_143_flush_to_zero',
          dtype=dtype,
          argument_result_values=[
              [fp8_143_min_value, fp8_143_min_value],
              [float.fromhex('0x1p-12'), 0.],
              [float.fromhex('0x1p-13'), 0.],
              [-fp8_143_min_value, -fp8_143_min_value],
              [float.fromhex('-0x1p-12'), -0.],
              [float.fromhex('-0x1p-13'), -0.],
          ],
      ),
      dict(
          testcase_name=f'{dtype.dtype.name}_fp8_143_round_near_even',
          dtype=dtype,
          argument_result_values=[
              [-1. + 1 / 32, -1.0],
              [-1. - 1 / 16, -1.0],
              [-2. + 1 / 16, -2.0],
              [-2. - 1 / 8, -2.0],
              [1. - 1 / 32, 1.0],
              [1. + 1 / 16, 1.0],
              [2. - 1 / 16, 2.0],
              [2. + 1 / 8, 2.0],
              [27.5, 28.],
              [28.5, 28.],
              [-27.5, -28.],
              [-28.5, -28.],
          ],
      ),
      dict(
          testcase_name=f'{dtype.dtype.name}_fp8_143_multi_dimensional',
          dtype=dtype,
          argument_result_values=[
              [
                  [
                      [
                          [7.09375, float('inf'), 8.125, 14.1875],
                          [48., 11.875, 8.5625, 9.3125],
                          [4.15625, 0.433594, 15.4375, 2.32812],
                      ],
                      [
                          [5.4375, 1.00781, 13.25, 1.57812],
                          [15.3125, 2.23438, 6.71875, 7.9375],
                          [3.0625, 6.6875, 4.625,
                           float('nan')],
                      ],
                  ],
                  [
                      [
                          [7, float('inf'), 8, 14],
                          [30., 12, 9, 9],
                          [4, 0.4375, 15, 2.25],
                      ],
                      [
                          [5.5, 1, 13, 1.625],
                          [15, 2.25, 6.5, 8],
                          [3, 6.5, 4.5, float('nan')],
                      ],
                  ],
              ],
          ],
      ),
  )


def gradient_test_data(sig_bits):
  fp_format = dict(exp_min=-11, exp_max=4, sig_bits=sig_bits)
  bounds = fp_cast.get_bounds(**fp_format)

  # Approximate float32 eps at bounds.flush_to_zero_bound
  eps_flush_to_zero = 1e-7
  # Approximate float32 eps at bounds.saturation_bound
  eps_saturation = 1e-5

  # Construct a list of pairs where the first element is the primal value
  # and the second element is the expected gradient of the floating-point
  # quantization function evaluated at that value.
  testcases = [

      # Test values above the saturation bound. Gradient should be zero since
      # these values are clipped to the saturation threshold during
      # quantization.
      (2.11e10, 0),
      (2.11e10, 0),
      (bounds.saturation_bound + eps_saturation, 0),
      (-bounds.saturation_bound - eps_saturation, 0),

      # Test values within the range of the fp format. The gradient should be 1
      # since values in this range use the straight-through estimator.
      (1.0, 1),
      (-1.0, 1),
      (2.11, 1),
      (-2.11, 1),
      (bounds.flush_to_zero_bound + eps_flush_to_zero, 1),
      (-bounds.flush_to_zero_bound - eps_flush_to_zero, 1),
      (bounds.saturation_bound - eps_saturation, 1),
      (-bounds.saturation_bound + eps_saturation, 1)
  ]
  return [{'primal': testcase[0], 'expected_grad': testcase[1], **fp_format
           } for testcase in testcases]


class FpCastTest(parameterized.TestCase):

  @parameterized.named_parameters(*test_data(jnp.bfloat16),
                                  *test_data(jnp.float32))
  def test_downcast_sat_ftz(self, dtype, argument_result_values):
    argument_result = jnp.array(
        argument_result_values,
        dtype=dtype,
    )
    y = fp_cast.downcast_sat_ftz(
        argument_result[:, 0],
        exp_min=-11,
        exp_max=4,
        sig_bits=3,
    )
    onp.testing.assert_equal(
        onp.array(argument_result[:, 1], dtype=onp.float32),
        onp.array(y, dtype=onp.float32),
    )

  def test_invalid_argument_type(self):
    x_s8 = jnp.array(1, dtype=jnp.int8)
    with self.assertRaises(ValueError):
      fp_cast.downcast_sat_ftz(x_s8, exp_min=-11, exp_max=4, sig_bits=3)

  def test_return_type(self):
    x_bf16 = jnp.array(1.0, dtype=jnp.bfloat16)
    y_bf16 = fp_cast.downcast_sat_ftz(
        x_bf16, exp_min=-11, exp_max=4, sig_bits=3)
    self.assertEqual(x_bf16.dtype, y_bf16.dtype)

    xf32 = jnp.array(1.0, dtype=jnp.bfloat16)
    yf32 = fp_cast.downcast_sat_ftz(xf32, exp_min=-11, exp_max=4, sig_bits=3)
    self.assertEqual(xf32.dtype, yf32.dtype)

  def test_sig_bits_zero(self):
    x = jnp.array(2.11111)
    y = fp_cast.downcast_sat_ftz(x, exp_min=-11, exp_max=4, sig_bits=0)
    self.assertEqual(y.item(), 2.0)

  @parameterized.parameters(*gradient_test_data(sig_bits=3),
                            *gradient_test_data(sig_bits=0))
  def test_grad(self, primal, expected_grad, sig_bits, exp_min, exp_max):
    # We use a 'sum' here so that each element of the gradient of the sum
    # corresponds to the partial derivative of output of `downcast_sat_ftz`
    # with respect to the corresponding input.
    def downcast_and_sum(x):
      return jnp.sum(
          fp_cast.downcast_sat_ftz(
              x, sig_bits=sig_bits, exp_min=exp_min, exp_max=exp_max))

    x_grad = jax.grad(downcast_and_sum)(jnp.array(primal))
    onp.testing.assert_equal(onp.array(x_grad), expected_grad)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized conv_general."""

import functools
from typing import Optional, Sequence, Tuple, Union

from aqt.common import aqt_config
from aqt.common import aqt_config_utils
from aqt.jax import aqt_tensor
from aqt.jax import aqt_utils
import jax
from jax import lax
import jax.numpy as jnp

# pylint: disable=protected-access
# pytype: disable=attribute-error


@functools.partial(
    jax.custom_jvp, nondiff_argnums=(2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12))
def _conv_general_aqt(
    lhs: jnp.ndarray,  #
    rhs: jnp.ndarray,
    lhs_quantizer: aqt_tensor.TensorQuantizer,
    rhs_quantizer: aqt_tensor.TensorQuantizer,
    should_int8_quantize: bool,
    train: bool,
    window_strides: Sequence[int],
    padding: Union[str, Sequence[Tuple[int, int]]],
    lhs_dilation: Optional[Sequence[int]],
    rhs_dilation: Optional[Sequence[int]],
    dimension_numbers: lax.ConvGeneralDilatedDimensionNumbers,
    feature_group_count: int,
    batch_group_count: int) -> jnp.ndarray:
  """Wrapper around lax.conv_general_dilated, but with option to use integer conv."""

  def conv_general_float(ops):
    lhs_, rhs_ = ops
    lhs_ = aqt_utils.possibly_use_quantized_variable(lhs_quantizer, lhs_, train)
    rhs_ = aqt_utils.possibly_use_quantized_variable(rhs_quantizer, rhs_, train)
    return lax.conv_general_dilated(
        lhs_,
        rhs_,
        window_strides=window_strides,
        padding=padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count)

  def conv_general_int(ops):
    lhs_, rhs_ = ops
    lhs_int = lhs_.astype(jnp.int8)
    rhs_int = rhs_.astype(jnp.int8)

    lhs_int = aqt_utils.possibly_use_quantized_variable(lhs_quantizer, lhs_int,
                                                        train)
    rhs_int = aqt_utils.possibly_use_quantized_variable(rhs_quantizer, rhs_int,
                                                        train)

    return lax.conv_general_dilated(
        lhs_int,
        rhs_int,
        window_strides=window_strides,
        padding=padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count,
        preferred_element_type=jnp.int32).astype(jnp.float32)

  return lax.cond(should_int8_quantize, conv_general_int, conv_general_float,
                  (lhs, rhs))


@_conv_general_aqt.defjvp
def _conv_general_aqt_jvp(
    lhs_quantizer: aqt_tensor.TensorQuantizer,  #
    rhs_quantizer: aqt_tensor.TensorQuantizer,
    should_int8_quantize: bool,
    train: bool,
    window_strides: Sequence[int],
    padding: Union[str, Sequence[Tuple[int, int]]],
    lhs_dilation: Optional[Sequence[int]],
    rhs_dilation: Optional[Sequence[int]],
    dimension_numbers: lax.ConvGeneralDilatedDimensionNumbers,
    feature_group_count: int,
    batch_group_count: int,
    primals,
    tangents) -> jnp.ndarray:
  """Custom gradient for conv_general_aqt that ignores integer casts."""
  lhs, rhs = primals
  lhs_dot, rhs_dot = tangents
  y = _conv_general_aqt(
      lhs,
      rhs,
      lhs_quantizer,
      rhs_quantizer,
      should_int8_quantize=should_int8_quantize,
      train=train,
      window_strides=window_strides,
      padding=padding,
      lhs_dilation=lhs_dilation,
      rhs_dilation=rhs_dilation,
      dimension_numbers=dimension_numbers,
      feature_group_count=feature_group_count,
      batch_group_count=batch_group_count)

  def differentiable_conv_general(lhs_, rhs_):
    # The backward pass op is performed in floating point for now and thus
    # here we use lax.conv_general_dilated to be differentiated by Autodiff.
    return lax.conv_general_dilated(
        lhs_,
        rhs_,
        window_strides=window_strides,
        padding=padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count)

  _, y_tangent = jax.jvp(
      differentiable_conv_general,  #
      (lhs, rhs),
      (lhs_dot, rhs_dot))
  return y, y_tangent  # pytype: disable=bad-return-type  # jax-ndarray


def _validate_dilation_argument(
    lhs_quantizer,  #
    rhs_quantizer,
    lhs_dilation,
    rhs_dilation):
  """Validates preserve_zero=True of AqtIntQuantConfig when the dilation is used."""
  lhs_configs = lhs_quantizer.config.tensor_configs
  rhs_configs = rhs_quantizer.config.tensor_configs
  event_count = lhs_quantizer._last_update.value

  for lhs_config, rhs_config in zip(lhs_configs, rhs_configs):
    if (isinstance(lhs_config.quant_config, aqt_config.IntQuantConfig) and
        isinstance(rhs_config.quant_config, aqt_config.IntQuantConfig) and
        lhs_config.quant_config.bits <= 8 and
        rhs_config.quant_config.bits <= 8):
      is_config_active = aqt_tensor.is_config_active(lhs_config, event_count)
      if is_config_active:
        if lhs_dilation is not None and not lhs_config.quant_config.preserve_zero:
          raise aqt_config.ConfigError(
              f'lhs_config.quant_config.preserve_zero ({lhs_config.quant_config.preserve_zero}) must be True if the input is dilated.'
          )
        if rhs_dilation is not None and not rhs_config.quant_config.preserve_zero:
          raise aqt_config.ConfigError(
              f'rhs_config.quant_config.preserve_zero ({rhs_config.quant_config.preserve_zero}) must be True if the filter is dilated.'
          )


def _validate_inputs(
    input_quantizer: aqt_tensor.TensorQuantizer,  #
    filter_quantizer: aqt_tensor.TensorQuantizer,
    dimension_numbers: lax.ConvGeneralDilatedDimensionNumbers) -> None:
  """Validates configs and inputs for conv_general_dilated.

  Args:
    input_quantizer: the input tensor quantizer.
    filter_quantizer: the filter tensor quantizer.
    dimension_numbers: the conv input tensor argument axes format.

  Raises:
    aqt_config.ConfigError: The input or filter quantizer configurations
      do not share statistics along contraction axes as described by
      `_validate_contraction`, or their quantization schedules are not aligned
      per `aqt_config._validate_alignment`.
  """

  if input_quantizer.config is not None and filter_quantizer.config is not None:
    aqt_config_utils._validate_alignment(
        'input_quantizer.config.tensor_configs',
        input_quantizer.config.tensor_configs,
        'filter_quantizer.config.tensor_configs',
        filter_quantizer.config.tensor_configs)

  input_spec, filter_spec, _ = dimension_numbers  # pytype: disable=attribute-error
  _, *input_contracted_dims = input_spec
  _, *filter_contracted_dims = filter_spec

  if input_quantizer.config is not None:
    for axis in input_contracted_dims:
      if axis not in input_quantizer.config.stats_config.share_stats_axes:
        raise aqt_config.ConfigError(
            f'expected contraction axis ({axis}) to be in '
            f'input_quantizer.config.stats_config.share_stats_axes={input_quantizer.config.stats_config.share_stats_axes}'
        )
  if filter_quantizer.config is not None:
    for axis in filter_contracted_dims:
      if axis not in filter_quantizer.config.stats_config.share_stats_axes:
        raise aqt_config.ConfigError(
            f'expected contraction axis ({axis}) to be in '
            f'filter_quantizer.config.stats_config.share_stats_axes={filter_quantizer.config.stats_config.share_stats_axes}'
        )


def _transpose_inv_scale(x, dimension_numbers_before, dimension_numbers_after):
  """Changes the order of axes in x from dimension_numbers_before to dimension_numbers_after."""
  assert (len(dimension_numbers_before) == len(dimension_numbers_after)), (
      f'len(dimension_numbers_before) ({len(dimension_numbers_before)}) must ',
      f'be equal to len(dimension_numbers_after) ({len(dimension_numbers_after)})'
  )

  axes = [0] * len(dimension_numbers_before)
  for i, axis in enumerate(dimension_numbers_after):
    axes[axis] = dimension_numbers_before[i]

  return jnp.transpose(x, axes)


def conv_general_dilated(lhs: jnp.ndarray,
                         rhs: jnp.ndarray,
                         lhs_quantizer: aqt_tensor.TensorQuantizer,
                         rhs_quantizer: aqt_tensor.TensorQuantizer,
                         window_strides: Sequence[int],
                         padding: Union[str, Sequence[Tuple[int, int]]],
                         lhs_dilation: Optional[Sequence[int]] = None,
                         rhs_dilation: Optional[Sequence[int]] = None,
                         dimension_numbers: Optional[
                             lax.ConvGeneralDilatedDimensionNumbers] = None,
                         feature_group_count: int = 1,
                         batch_group_count: int = 1,
                         train: bool = True) -> jnp.ndarray:
  """Quantized jax.lax.conv_general_dilated.

  Args:
    lhs: Left-hand side of the conv_general_dilated (input).
    rhs: Right-hand side of the conv_general_dilated (filter).
    lhs_quantizer: The tensor quantizer for lhs.
    rhs_quantizer: The tensor quantizer for rhs.
    window_strides: a sequence of `n` integers, representing the inter-window
      strides.
    padding: either the string `'SAME'`, the string `'VALID'`, or a sequence of
      `n` `(low, high)` integer pairs that give the padding to apply before and
      after each spatial dimension.
    lhs_dilation: `None`, or a sequence of `n` integers, giving the dilation
      factor to apply in each spatial dimension of `lhs`. LHS dilation is also
      known as transposed convolution.
    rhs_dilation: `None`, or a sequence of `n` integers, giving the dilation
      factor to apply in each spatial dimension of `rhs`. RHS dilation is also
      known as atrous convolution.
    dimension_numbers: either `None`, a ``ConvDimensionNumbers`` object, or a
      3-tuple ``(lhs_spec, rhs_spec, out_spec)``, where each element is a string
      of length `n+2`.
    feature_group_count: integer, default 1. See XLA HLO docs.
    batch_group_count: integer, default 1. See XLA HLO docs.
    train: If false and `use_quantized_variable` in lhs_quantizer or
      rhs_quantizer, then this indicates `aqt_conv_general` should use the
      quantized variable with the latest quantized, memorized from the most
      recent `TensorQuantizer.update()` in quantized operations rather than the
      float tensor input `lhs` or `rhs` provided to those operations at
      inference time.

  Returns:
    An array containing the result with the same dtype as 'lhs' and 'rhs'.
  """
  # TODO(jihwanlee): Support quantization for non-2D convolution.
  if len(lhs.shape) != 4:
    raise NotImplementedError('Currently, aqt_conv_general_dilated supports ',
                              'only 2-D convolution.')

  dimension_numbers = lax.conv_dimension_numbers(lhs.shape, rhs.shape,
                                                 dimension_numbers)

  _validate_inputs(lhs_quantizer, rhs_quantizer, dimension_numbers)

  lhs_scale, lhs_inv_scale = lhs_quantizer._get_quant_scale(train)
  rhs_scale, rhs_inv_scale = rhs_quantizer._get_quant_scale(train)

  lhs = lhs_scale * lhs
  rhs = rhs_scale * rhs

  lhs = lhs_quantizer._to_quant(lhs, train)
  rhs = rhs_quantizer._to_quant(rhs, train)

  should_int8_quantize = aqt_utils.should_int8_quantize(lhs_quantizer,
                                                        rhs_quantizer)

  if should_int8_quantize:
    _validate_dilation_argument(lhs_quantizer, rhs_quantizer, lhs_dilation,
                                rhs_dilation)

  conv = _conv_general_aqt(lhs, rhs, lhs_quantizer, rhs_quantizer,
                           should_int8_quantize, train, window_strides, padding,
                           lhs_dilation, rhs_dilation, dimension_numbers,
                           feature_group_count, batch_group_count)

  # Transpose both lhs_inv_scale and rhs_inv_scale such that they have `NHWC`
  # and `HWIO` shapes, respectively, which allows them to be broadcastable no
  # matter how they were shaped originally by `dimension_numbers`.
  dim_numbers = {
      'NHWC': (0, 3, 1, 2),
      'HWIO': (3, 2, 0, 1),
  }
  if lhs_quantizer.config is not None and rhs_quantizer.config is not None:
    lhs_inv_scale = _transpose_inv_scale(lhs_inv_scale, dimension_numbers[0],
                                         dim_numbers['NHWC'])
    rhs_inv_scale = _transpose_inv_scale(rhs_inv_scale, dimension_numbers[1],
                                         dim_numbers['HWIO'])
    assert len(lhs_inv_scale.shape) == len(rhs_inv_scale.shape)

  inv_scale = lhs_inv_scale * rhs_inv_scale
  # Reverse the shape of inv_scale back to one specified by out_spec in
  # `dimension_numbers`
  inv_scale = _transpose_inv_scale(inv_scale, dim_numbers['NHWC'],
                                   dimension_numbers[2])
  return conv * inv_scale



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized dot_general."""

import functools

from aqt.common import aqt_config
from aqt.common import aqt_config_utils
from aqt.jax import aqt_tensor
from aqt.jax import aqt_utils
import jax
from jax import lax
import jax.numpy as jnp

# TODO(b/220181240): Remove accesses to protected methods. e.g., TQ._to_quant()
# -> TQ.to_quant().
# We repeatedly use protected methods from classes defined in other modules to
# avoid exporting them as part of the public API.
# pylint: disable=protected-access


@functools.partial(jax.custom_jvp, nondiff_argnums=(2, 3, 4, 5, 6))
def _dot_general_aqt(
    lhs,  #
    rhs,
    lhs_quantizer,
    rhs_quantizer,
    dimension_numbers,
    should_int8_quantize,
    train):
  """Wrapper around lax.dot_general, but with option to use integer dot.

  This function comes equipped with a custom gradient that defines the
  gradient of this function to be the same as the equivalent call to
  lax.dot_general, ignoring casts to and from integer types so that
  quantization-aware-training will work correctly.

  See docstring of lax.dot_general.

  Args:
    lhs: Left-hand side of the dot_general.
    rhs: Right-hand side of the dot_general.
    lhs_quantizer: The tensor quantizer for lhs.
    rhs_quantizer: The tensor quantizer for rhs.
    dimension_numbers: a tuple of tuples of the form `((lhs_contracting_dims,
      rhs_contracting_dims), (lhs_batch_dims, rhs_batch_dims))`
    should_int8_quantize: If true, inputs to lax.dot_general will be cast to
      int8 and results accumulated to int32, then converted back to the original
      input type.
    train: If false and `TensorQuantizer.use_quantized_variable` is True, then
      use the quantized variable, instead of input tensors, for the respective
      input tensor.

  Returns:
    Same as lax.dot_general, but its quantized version.
  """

  def dot_general_float(ops):
    lhs_, rhs_ = ops
    lhs_ = aqt_utils.possibly_use_quantized_variable(lhs_quantizer, lhs_, train)
    rhs_ = aqt_utils.possibly_use_quantized_variable(rhs_quantizer, rhs_, train)
    return lax.dot_general(lhs_, rhs_, dimension_numbers=dimension_numbers)

  def dot_general_int(ops):
    lhs_, rhs_ = ops
    lhs_int = lhs_.astype(jnp.int8)
    rhs_int = rhs_.astype(jnp.int8)

    lhs_int = aqt_utils.possibly_use_quantized_variable(lhs_quantizer, lhs_int,
                                                        train)
    rhs_int = aqt_utils.possibly_use_quantized_variable(rhs_quantizer, rhs_int,
                                                        train)

    return lax.dot_general(
        lhs_int,
        rhs_int,
        dimension_numbers=dimension_numbers,
        preferred_element_type=jnp.int32).astype(jnp.float32)

  return lax.cond(should_int8_quantize, dot_general_int, dot_general_float,
                  (lhs, rhs))


@_dot_general_aqt.defjvp
def _dot_general_aqt_jvp(
    lhs_quantizer,  #
    rhs_quantizer,
    dimension_numbers,
    should_int8_quantize,
    train,
    primals,
    tangents):
  """Custom gradient for dot_general_aqt that ignores integer casts."""
  lhs, rhs = primals
  lhs_dot, rhs_dot = tangents
  y = _dot_general_aqt(
      lhs,
      rhs,
      lhs_quantizer,
      rhs_quantizer,
      dimension_numbers=dimension_numbers,
      should_int8_quantize=should_int8_quantize,
      train=train)

  def differentiable_dot_general(lhs_, rhs_):
    return lax.dot_general(lhs_, rhs_, dimension_numbers=dimension_numbers)

  _, y_tangent = jax.jvp(
      differentiable_dot_general,  #
      (lhs, rhs),
      (lhs_dot, rhs_dot))
  return y, y_tangent


def _validate_inputs(
    lhs_quantizer: aqt_tensor.TensorQuantizer,  #
    rhs_quantizer: aqt_tensor.TensorQuantizer,
    dimension_numbers: lax.DotDimensionNumbers):
  """Validates configs and inputs for dot_general."""

  lhs_config = lhs_quantizer.config
  rhs_config = rhs_quantizer.config

  if lhs_config is not None and rhs_config is not None:
    aqt_config_utils._validate_alignment(
        'lhs_config',  #
        lhs_config.tensor_configs,
        'rhs_config',
        rhs_config.tensor_configs)

  lhs_contracting_dims, rhs_contracting_dims = dimension_numbers[0]

  if not (lhs_config is None or
          any(a in lhs_config.stats_config.share_stats_axes
              for a in lhs_contracting_dims)):
    raise aqt_config.ConfigError(
        f'expected lhs dot_general contraction axis to be in '
        f'share_stats_axes={lhs_config.stats_config.share_stats_axes}')
  if not (rhs_config is None or
          any(a in rhs_config.stats_config.share_stats_axes
              for a in rhs_contracting_dims)):
    raise aqt_config.ConfigError(
        f'expected rhs dot_general contraction axis to be in '
        f'share_stats_axes={rhs_config.stats_config.share_stats_axes}')


def dot(lhs: jnp.ndarray, rhs: jnp.ndarray,
        lhs_quantizer: aqt_tensor.TensorQuantizer,
        rhs_quantizer: aqt_tensor.TensorQuantizer) -> jnp.ndarray:
  """Quantized lax.dot.

  This dot operator is based on lax.dot,
  https://jax.readthedocs.io/en/latest/_modules/jax/_src/lax/lax.html#dot.
  For more general contraction, see the `dot_general` operator.

  Args:
    lhs: Left-hand side of the dot.
    rhs: Right-hand side of the dot.
    lhs_quantizer: The tensor quantizer for lhs.
    rhs_quantizer: The tensor quantizer for rhs.

  Returns:
    An array containing the result with the same dtype as 'lhs' and 'rhs'.
  """
  if 1 <= lhs.ndim <= 2 and 1 <= rhs.ndim <= 2 and lhs.shape[-1] == rhs.shape[0]:
    return dot_general(
        lhs,
        rhs,
        lhs_quantizer,
        rhs_quantizer,
        dimension_numbers=(((lhs.ndim - 1,), (0,)), ((), ())))
  else:
    raise TypeError('Incompatible shapes for dot: got {} and {}.'.format(
        lhs.shape, rhs.shape))


def dot_general(lhs: jnp.ndarray,
                rhs: jnp.ndarray,
                lhs_quantizer: aqt_tensor.TensorQuantizer,
                rhs_quantizer: aqt_tensor.TensorQuantizer,
                dimension_numbers: lax.DotDimensionNumbers,
                train: bool = True) -> jnp.ndarray:
  """Quantized jax.lax.dot_general.

  Args:
    lhs: Left-hand side of the dot_general.
    rhs: Left-hand side of the dot_general.
    lhs_quantizer: The tensor quantizer for lhs.
    rhs_quantizer: The tensor quantizer for rhs.
    dimension_numbers: a tuple of tuples of the form `((lhs_contracting_dims,
      rhs_contracting_dims), (lhs_batch_dims, rhs_batch_dims))`
    train: If false and `use_quantized_variable` in lhs_quantizer or
      rhs_quantizer, then this indicates `aqt_dot_general` should use the
      quantized variable with the latest quantized, memorized from the most
      recent `TensorQuantizer.update()` in quantized operations rather than the
      float tensor input `lhs` or `rhs` provided to those operations at
      inference time.

  Returns:
    An array containing the result with the same dtype as 'lhs' and 'rhs'.
  """
  _validate_inputs(lhs_quantizer, rhs_quantizer, dimension_numbers)

  lhs_scale, lhs_inv_scale = lhs_quantizer._get_quant_scale(train)
  rhs_scale, rhs_inv_scale = rhs_quantizer._get_quant_scale(train)

  lhs = lhs_scale * lhs
  rhs = rhs_scale * rhs

  lhs = lhs_quantizer._to_quant(lhs, train)
  rhs = rhs_quantizer._to_quant(rhs, train)

  should_int8_quantize = aqt_utils.should_int8_quantize(lhs_quantizer,
                                                        rhs_quantizer)

  out = _dot_general_aqt(
      lhs,
      rhs,
      lhs_quantizer,
      rhs_quantizer,
      dimension_numbers=dimension_numbers,
      should_int8_quantize=should_int8_quantize,
      train=train)

  if lhs_quantizer.config is not None and rhs_quantizer.config is not None:
    inv_scale = lax.dot_general(
        lhs_inv_scale, rhs_inv_scale, dimension_numbers=dimension_numbers)
  else:
    inv_scale = lhs_inv_scale * rhs_inv_scale

  return out * inv_scale


def injectable_dot_general(lhs_quantizer, rhs_quantizer, train):
  """Wrapper of aqt_dot_general, supposed to be used in injection API."""
  return lambda a, b, dimension_numbers, precision: dot_general(  # pylint: disable=g-long-lambda
      lhs=a,
      rhs=b,
      lhs_quantizer=lhs_quantizer,
      rhs_quantizer=rhs_quantizer,
      dimension_numbers=dimension_numbers,
      train=train)
